import random

from .run_monte_carlo_diffusion import run_monte_carlo_diffusion
from ..graph import IMGraph
from .base_diffusion_model import BaseDiffusionModel
from ...utils import infection_threshold


class SusceptibleInfectedRecoveredModel(BaseDiffusionModel):
    """
    SIR模型（Susceptible-Infected-Recovered Model）。

    实现了经典的SIR传播模型，该模型是一种传染病传播模型，
    节点有三种状态：易感(S)、感染(I)和康复(R)。
    易感节点可被感染节点感染，感染节点可以康复为免疫状态。

    参考文献:
        - Kermack, W. O., & McKendrick, A. G. (1927). "A contribution to the mathematical theory of epidemics."
          Proceedings of the Royal Society of London. Series A, 115(772), 700-721.
          DOI: 10.1098/rspa.1927.0118
          URL: https://royalsocietypublishing.org/doi/10.1098/rspa.1927.0118

    Attributes:
        infected_nodes (set): 所有已被感染的节点集合
        recovered_nodes (set): 所有已康复的节点集合
        graph (IMGraph): 表示传播网络结构（继承自BaseDiffusionModel）
        init_seeds (list): 初始感染节点集（继承自BaseDiffusionModel）
        record_states (bool): 指示是否记录传播过程中的状态变化（继承自BaseDiffusionModel）
        states (list): 当record_states为True时存储传播过程的状态历史（继承自BaseDiffusionModel）
        gamma (float): 康复率
        beta (float): 感染率
    """

    def __init__(self, graph: IMGraph, init_seeds: list, gamma: float, beta: float = None,
                 record_states: bool = False):
        """
        初始化SIR模型。

        Args:
            graph (IMGraph): 定义了节点和边的网络结构
            init_seeds (list): 初始感染的节点集合
            gamma (float): 康复率，每个时间步感染节点康复的概率
            beta (float, optional): 感染率，如果为None则使用基于图结构计算的感染阈值
            record_states (bool): 控制是否记录每一步的状态，默认为False
        """
        super(SusceptibleInfectedRecoveredModel, self).__init__(graph, init_seeds, record_states)
        self.infected_nodes = set(init_seeds)
        self.recovered_nodes = set()
        self.gamma = gamma
        if beta is None:
            self.beta = infection_threshold(self.graph.nx_graph)
        else:
            self.beta = beta

    def update(self):
        """
        执行一次传播更新。

        在当前轮次中，先处理感染节点的康复，然后已感染的节点尝试感染它们的易感邻居节点。

        Returns:
            set: 本轮新感染的节点集合
        """
        # 先执行恢复过程 - 所有感染节点都有一定概率康复
        new_recovered_nodes = set()
        for node in self.infected_nodes:
            if random.random() < self.gamma:
                new_recovered_nodes.add(node)
                self.recovered_nodes.add(node)

        # 从感染节点集合中移除已康复的节点
        self.infected_nodes -= new_recovered_nodes

        # 再执行感染过程 - 所有当前感染节点尝试感染邻居
        new_infected_nodes = set()
        for node in list(self.infected_nodes):  # 使用list复制避免在迭代时修改集合
            for neighbor in self.graph.neighbors(node):
                # 只有易感节点（既未感染也未康复）才能被感染
                if neighbor not in self.infected_nodes and neighbor not in self.recovered_nodes:
                    if random.random() < self.beta:  # 使用beta作为感染概率
                        new_infected_nodes.add(neighbor)

        self.infected_nodes.update(new_infected_nodes)
        # 记录每轮的状态
        if self.record_states and (new_infected_nodes or new_recovered_nodes):
            self.states.append({
                'newly_infected': new_infected_nodes.copy(),
                'newly_recovered': new_recovered_nodes.copy()
            })

        return new_infected_nodes

    def diffusion(self, update_counts=None):
        """
        执行完整的扩散过程。

        进行指定轮次的传播更新，直到达到最大轮次或没有新的节点被感染。

        Args:
            update_counts (int, optional): 最大更新轮次数。如果为None则持续传播直到无新节点感染

        Returns:
            dict: 包含最终各状态节点集合的字典
                - 'infected': 最终被感染的节点集合
                - 'recovered': 最终已康复的节点集合
                - 'susceptible': 最终仍处于易感状态的节点集合
        """
        if update_counts is not None and update_counts <= 0:
            raise ValueError("update_counts must be a positive integer.")

        count = 0
        while True:
            _ = self.update()
            count += 1
            # 当没有新的感染节点或者达到指定轮次时停止
            if len(self.infected_nodes) == 0 or (update_counts and count >= update_counts):
                break

        return self.recovered_nodes

    def run_monte_carlo_diffusion(self, rounds: int, update_counts: int = None, multi_process: bool = False,
                                  processes: int = None, seed: int = None):
        """
        执行蒙特卡洛模拟扩散过程。

        Args:
            rounds (int): 总模拟轮数
            update_counts (int, optional): 更新轮次数，适用于SI模型等需要限制传播轮次的模型
            multi_process (bool): 是否启用多进程模式，默认为False
            processes (int, optional): 多进程模式下的进程数，为None时使用CPU核心数
            seed (int, optional): 模拟时的随机种子

        Returns:
            float: 所有模拟轮次的平均激活节点数
        """
        return run_monte_carlo_diffusion(self, rounds, update_counts, multi_process, processes, seed)


    def reset(self, init_seeds=None):
        """
        重置模型状态。

        将模型恢复到初始状态，如果启用状态记录则清空已记录的状态历史，
        并重新设置初始感染节点状态。

        Args:
            init_seeds (list, optional): 新的初始感染节点集合，若为None则使用原有种子集
        """
        if init_seeds is None:
            init_seeds = self.init_seeds
        self.init_seeds = init_seeds
        self.infected_nodes = set(init_seeds)
        self.recovered_nodes = set()
        if self.record_states:
            self.states = []



import random
from pynetim.cpp.graph import IMGraphCpp

def test_unified_batch_interfaces():
    print("=" * 60)
    print("Testing Unified Batch Interfaces")
    print("=" * 60)
    
    num_nodes = 1000
    num_edges = 5000
    
    random.seed(42)
    edges = []
    weights = []
    for _ in range(num_edges):
        u = random.randint(0, num_nodes - 1)
        v = random.randint(0, num_nodes - 1)
        if u != v:
            edges.append((u, v))
            weights.append(random.random())
    
    print(f"\nCreating graph with {num_nodes} nodes and {len(edges)} edges...")
    graph = IMGraphCpp(num_nodes=num_nodes, edges=edges, weights=weights, directed=True)
    print(f"Graph: {graph}")
    
    test_nodes = list(range(0, 100, 5))
    
    print("\n" + "=" * 60)
    print("Test 1: batch_out_degree")
    print("=" * 60)
    out_degrees = graph.batch_out_degree(test_nodes)
    print(f"Test nodes: {len(test_nodes)}")
    print(f"Sample results (first 5):")
    for i in range(min(5, len(test_nodes))):
        print(f"  Node {test_nodes[i]}: out_degree = {out_degrees[i]}")
    
    print("\n" + "=" * 60)
    print("Test 2: batch_in_degree")
    print("=" * 60)
    in_degrees = graph.batch_in_degree(test_nodes)
    print(f"Test nodes: {len(test_nodes)}")
    print(f"Sample results (first 5):")
    for i in range(min(5, len(test_nodes))):
        print(f"  Node {test_nodes[i]}: in_degree = {in_degrees[i]}")
    
    print("\n" + "=" * 60)
    print("Test 3: batch_degree")
    print("=" * 60)
    degrees = graph.batch_degree(test_nodes)
    print(f"Test nodes: {len(test_nodes)}")
    print(f"Sample results (first 5):")
    for i in range(min(5, len(test_nodes))):
        print(f"  Node {test_nodes[i]}: degree = {degrees[i]}")
        print(f"    (out_degree={out_degrees[i]}, in_degree={in_degrees[i]}, total={degrees[i]})")
    
    print("\n" + "=" * 60)
    print("Test 4: batch_out_neighbors")
    print("=" * 60)
    neighbors = graph.batch_out_neighbors(test_nodes[:5])
    print(f"Test nodes: 5")
    for i, node in enumerate(test_nodes[:5]):
        print(f"  Node {node}: {len(neighbors[i])} neighbors")
        if neighbors[i]:
            print(f"    First neighbor: ({neighbors[i][0][0]}, weight={neighbors[i][0][1]:.4f})")
    
    print("\n" + "=" * 60)
    print("Test 5: batch_get_edge_weight")
    print("=" * 60)
    test_edges = [(test_nodes[i], test_nodes[(i+1) % len(test_nodes)]) for i in range(10)]
    edge_weights = graph.batch_get_edge_weight(test_edges)
    print(f"Test edges: {len(test_edges)}")
    print(f"Sample results:")
    for i, (u, v) in enumerate(test_edges[:5]):
        print(f"  Edge ({u}, {v}): weight = {edge_weights[i]:.4f}")
    
    print("\n" + "=" * 60)
    print("All unified batch interface tests passed!")
    print("=" * 60)

if __name__ == "__main__":
    test_unified_batch_interfaces()



import time
import random
from pynetim.cpp.graph import IMGraphCpp

def test_batch_operations():
    print("=" * 60)
    print("Testing Batch Operations")
    print("=" * 60)
    
    num_nodes = 1000
    num_edges = 5000
    
    random.seed(42)
    
    edges = []
    weights = []
    for _ in range(num_edges):
        u = random.randint(0, num_nodes - 1)
        v = random.randint(0, num_nodes - 1)
        if u != v:
            edges.append((u, v))
            weights.append(random.random())
    
    print(f"\nCreating graph with {num_nodes} nodes and {len(edges)} edges...")
    graph = IMGraphCpp(num_nodes=num_nodes, edges=edges, weights=weights, directed=True)
    print(f"Graph: {graph}")
    
    print("\n" + "=" * 60)
    print("Test 1: batch_out_degree")
    print("=" * 60)
    
    test_nodes = list(range(0, 100, 5))
    
    start_time = time.time()
    individual_degrees = [graph.out_degree(u) for u in test_nodes]
    individual_time = time.time() - start_time
    
    start_time = time.time()
    batch_degrees = graph.batch_out_degree(test_nodes)
    batch_time = time.time() - start_time
    
    print(f"Test nodes: {len(test_nodes)}")
    print(f"Individual query time: {individual_time:.6f}s")
    print(f"Batch query time: {batch_time:.6f}s")
    print(f"Speedup: {individual_time / batch_time:.2f}x")
    print(f"Results match: {individual_degrees == batch_degrees}")
    
    print("\n" + "=" * 60)
    print("Test 2: batch_out_neighbors")
    print("=" * 60)
    
    test_nodes = list(range(0, 50, 5))
    
    start_time = time.time()
    individual_neighbors = [graph.out_neighbors(u) for u in test_nodes]
    individual_time = time.time() - start_time
    
    start_time = time.time()
    batch_neighbors = graph.batch_out_neighbors(test_nodes)
    batch_time = time.time() - start_time
    
    print(f"Test nodes: {len(test_nodes)}")
    print(f"Individual query time: {individual_time:.6f}s")
    print(f"Batch query time: {batch_time:.6f}s")
    print(f"Speedup: {individual_time / batch_time:.2f}x")
    
    print(f"\nSample results (first 3 nodes):")
    for i in range(min(3, len(test_nodes))):
        u = test_nodes[i]
        print(f"  Node {u}: {len(individual_neighbors[i])} neighbors")
    
    print("\n" + "=" * 60)
    print("Test 3: batch_get_edge_weight")
    print("=" * 60)
    
    test_edges = edges[:100]
    
    start_time = time.time()
    individual_weights = [graph.get_edge_weight(u, v) for u, v in test_edges]
    individual_time = time.time() - start_time
    
    start_time = time.time()
    batch_weights = graph.batch_get_edge_weight(test_edges)
    batch_time = time.time() - start_time
    
    print(f"Test edges: {len(test_edges)}")
    print(f"Individual query time: {individual_time:.6f}s")
    print(f"Batch query time: {batch_time:.6f}s")
    print(f"Speedup: {individual_time / batch_time:.2f}x")
    print(f"Results match: {individual_weights == batch_weights}")
    
    print(f"\nSample weights (first 5 edges):")
    for i in range(min(5, len(test_edges))):
        u, v = test_edges[i]
        print(f"  Edge ({u}, {v}): {individual_weights[i]:.4f}")
    
    print("\n" + "=" * 60)
    print("Test 4: Large scale batch operations")
    print("=" * 60)
    
    large_nodes = list(range(num_nodes))
    
    start_time = time.time()
    all_degrees = graph.batch_out_degree(large_nodes)
    batch_time = time.time() - start_time
    
    print(f"Querying all {num_nodes} nodes...")
    print(f"Batch query time: {batch_time:.6f}s")
    print(f"Average degree: {sum(all_degrees) / num_nodes:.2f}")
    print(f"Max degree: {max(all_degrees)}")
    print(f"Min degree: {min(all_degrees)}")
    
    print("\n" + "=" * 60)
    print("All batch operation tests passed!")
    print("=" * 60)

if __name__ == "__main__":
    test_batch_operations()



import time
from functools import wraps


class Timer:
    """
    计时器装饰器类，用于测量函数执行时间。

    该装饰器可以重复运行函数多次并计算平均执行时间。

    Attributes:
        repeat (int): 函数每次调用时实际运行的次数
        total_time (float): 累计总执行时间
        call_count (int): 函数被调用的次数
    """

    def __init__(self, repeat=1):
        """
        初始化计时器装饰器。

        Args:
            repeat (int, optional): 函数每次调用时实际运行的次数，默认为1
        """
        self.repeat = repeat
        self.total_time = 0
        self.call_count = 0

    def __call__(self, func):
        """
        使Timer实例可调用，作为装饰器使用。

        Args:
            func (Callable): 被装饰的函数

        Returns:
            Callable: 包装后的函数
        """
        @wraps(func)
        def wrapper(*args, **kwargs):
            """
            包装函数，执行实际的计时逻辑。

            Args:
                *args: 传递给被装饰函数的位置参数
                **kwargs: 传递给被装饰函数的关键字参数

            Returns:
                Any: 被装饰函数的返回值
            """
            result = None
            for _ in range(self.repeat):
                start_time = time.time()
                result = func(*args, **kwargs)
                elapsed = time.time() - start_time
                self.total_time += elapsed
                self.call_count += 1
                # print(f"[Timer] {func.__name__} run {_ + 1}/{self.repeat}: {elapsed:.6f}s")
            print(f"运行时间: {average_time():.6f}s")
            return result

        def average_time():
            """
            计算函数的平均执行时间。

            Returns:
                float: 平均执行时间（秒）
            """
            if self.call_count == 0:
                return 0
            return self.total_time / self.call_count

        wrapper.average_time = average_time
        return wrapper



import random
import networkx as nx
from pynetim.py.graph import IMGraph as PyIMGraph
from pynetim.cpp.graph import IMGraphCpp

def test_interface_consistency():
    print("=" * 60)
    print("Testing Python/C++ Interface Consistency")
    print("=" * 60)
    
    num_nodes = 100
    num_edges = 500
    
    random.seed(42)
    
    nx_graph = nx.DiGraph()
    for _ in range(num_edges):
        u = random.randint(0, num_nodes - 1)
        v = random.randint(0, num_nodes - 1)
        if u != v:
            weight = random.random()
            nx_graph.add_edge(u, v, weight=weight)
    
    py_graph = PyIMGraph(nx_graph, edge_weight_type='WC')
    
    edges_list = [(u, v) for u, v in nx_graph.edges()]
    weights_list = [nx_graph[u][v]['weight'] for u, v in nx_graph.edges()]
    
    cpp_graph = IMGraphCpp(num_nodes=num_nodes, edges=edges_list, weights=weights_list, directed=True)
    
    test_nodes = list(range(0, 50, 5))
    
    print("\n" + "=" * 60)
    print("Test 1: batch_out_degree")
    print("=" * 60)
    py_out_degrees = py_graph.batch_out_degree(test_nodes)
    cpp_out_degrees = cpp_graph.batch_out_degree(test_nodes)
    
    print(f"Python result: {py_out_degrees[:5]}")
    print(f"C++ result: {cpp_out_degrees[:5]}")
    print(f"Results match: {py_out_degrees == cpp_out_degrees}")
    
    print("\n" + "=" * 60)
    print("Test 2: batch_in_degree")
    print("=" * 60)
    py_in_degrees = py_graph.batch_in_degree(test_nodes)
    cpp_in_degrees = cpp_graph.batch_in_degree(test_nodes)
    
    print(f"Python result: {py_in_degrees[:5]}")
    print(f"C++ result: {cpp_in_degrees[:5]}")
    print(f"Results match: {py_in_degrees == cpp_in_degrees}")
    
    print("\n" + "=" * 60)
    print("Test 3: batch_degree")
    print("=" * 60)
    py_degrees = py_graph.batch_degree(test_nodes)
    cpp_degrees = cpp_graph.batch_degree(test_nodes)
    
    print(f"Python result: {py_degrees[:5]}")
    print(f"C++ result: {cpp_degrees[:5]}")
    print(f"Results match: {py_degrees == cpp_degrees}")
    
    print("\n" + "=" * 60)
    print("Test 4: batch_out_neighbors")
    print("=" * 60)
    py_neighbors = py_graph.batch_out_degree(test_nodes[:5])
    cpp_neighbors = cpp_graph.batch_out_neighbors(test_nodes[:5])
    
    print(f"Python degrees: {py_neighbors}")
    print(f"C++ neighbor counts: {[len(n) for n in cpp_neighbors]}")
    
    print("\n" + "=" * 60)
    print("Test 5: batch_get_edge_weight")
    print("=" * 60)
    test_edges = [(test_nodes[i], test_nodes[(i+1) % len(test_nodes)]) for i in range(10)]
    
    py_weights = []
    cpp_weights = cpp_graph.batch_get_edge_weight(test_edges)
    
    for u, v in test_edges:
        if nx_graph.has_edge(u, v):
            py_weights.append(nx_graph[u][v]['weight'])
        else:
            py_weights.append(0.0)
    
    print(f"Python result: {py_weights[:5]}")
    print(f"C++ result: {cpp_weights[:5]}")
    print(f"Results match: {py_weights == cpp_weights}")
    
    print("\n" + "=" * 60)
    print("Interface consistency tests completed!")
    print("=" * 60)

if __name__ == "__main__":
    test_interface_consistency()



from .heuristic_algorithm import SingleDiscountAlgorithm, DegreeDiscountAlgorithm
from .simulation_algorithm import GreedyAlgorithm, CELFAlgorithm
from .RIS_algorithm import BaseRISAlgorithm, IMMAlgorithm
from .base_algorithm import BaseAlgorithm


__all__ = [
    'SingleDiscountAlgorithm',
    'DegreeDiscountAlgorithm',
    'GreedyAlgorithm',
    'CELFAlgorithm',
    'BaseRISAlgorithm',
    'IMMAlgorithm'
]



import psutil
import os
import time
from typing import Dict, Any, Optional

class MemoryMonitor:
    def __init__(self):
        self.process = psutil.Process(os.getpid())
        self.baseline_memory = self.get_memory_info()['rss']
        self.peak_memory = self.baseline_memory
    
    def get_memory_info(self) -> Dict[str, Any]:
        mem_info = self.process.memory_info()
        return {
            'rss': mem_info.rss / 1024 / 1024,
            'vms': mem_info.vms / 1024 / 1024,
            'shared': mem_info.shared / 1024 / 1024,
            'text': mem_info.text / 1024 / 1024,
            'lib': mem_info.lib / 1024 / 1024,
            'data': mem_info.data / 1024 / 1024,
            'dirty': mem_info.dirty / 1024 / 1024
        }
    
    def get_memory_usage(self) -> float:
        mem_info = self.get_memory_info()
        current_rss = mem_info['rss']
        self.peak_memory = max(self.peak_memory, current_rss)
        return current_rss
    
    def get_memory_increase(self) -> float:
        current_memory = self.get_memory_usage()
        return current_memory - self.baseline_memory
    
    def get_peak_memory(self) -> float:
        return self.peak_memory
    
    def reset_baseline(self):
        self.baseline_memory = self.get_memory_usage()
        self.peak_memory = self.baseline_memory
    
    def __enter__(self):
        self.reset_baseline()
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

def profile_memory_usage(func, *args, **kwargs) -> Dict[str, Any]:
    monitor = MemoryMonitor()
    start_time = time.time()
    
    result = func(*args, **kwargs)
    
    end_time = time.time()
    elapsed_time = end_time - start_time
    
    memory_info = monitor.get_memory_info()
    memory_increase = monitor.get_memory_increase()
    peak_memory = monitor.get_peak_memory()
    
    return {
        'result': result,
        'elapsed_time': elapsed_time,
        'memory_info': memory_info,
        'memory_increase_mb': memory_increase,
        'peak_memory_mb': peak_memory
    }

def format_memory_size(size_mb: float) -> str:
    if size_mb < 1024:
        return f"{size_mb:.2f} MB"
    else:
        return f"{size_mb / 1024:.2f} GB"

def print_memory_stats(stats: Dict[str, Any], label: str = ""):
    if label:
        print(f"\n{'='*60}")
        print(f"{label}")
        print(f"{'='*60}")
    
    print(f"Elapsed time: {stats['elapsed_time']:.4f} seconds")
    print(f"Memory increase: {format_memory_size(stats['memory_increase_mb'])}")
    print(f"Peak memory: {format_memory_size(stats['peak_memory_mb'])}")
    print(f"\nDetailed memory info:")
    for key, value in stats['memory_info'].items():
        print(f"  {key}: {format_memory_size(value)}")

if __name__ == "__main__":
    print("Memory Monitor Test")
    print("=" * 60)
    
    monitor = MemoryMonitor()
    print(f"Baseline memory: {format_memory_size(monitor.get_memory_usage())}")
    
    data = [i for i in range(1000000)]
    print(f"After creating list: {format_memory_size(monitor.get_memory_usage())}")
    print(f"Memory increase: {format_memory_size(monitor.get_memory_increase())}")
    
    del data
    print(f"After deleting list: {format_memory_size(monitor.get_memory_usage())}")
    
    print("\nTesting with context manager:")
    with MemoryMonitor() as m:
        large_data = [i for i in range(5000000)]
        time.sleep(0.1)
        print(f"Peak memory: {format_memory_size(m.get_peak_memory())}")
        print(f"Memory increase: {format_memory_size(m.get_memory_increase())}")



from typing import Set, TYPE_CHECKING

if TYPE_CHECKING:
    from ..graph import IMGraphCpp

class LinearThresholdModel:
    def __init__(self, graph: 'IMGraphCpp', seeds: Set[int], theta_l: float = 0.0, theta_h: float = 1.0, record_activated: bool = False) -> None: ...
    
    def set_seeds(self, new_seeds: Set[int]) -> None: ...
    
    def set_record_activated(self, record: bool) -> None: ...
    
    def run_single_simulation(self, seed: int = 0) -> Set[int]: ...
    
    def run_monte_carlo_diffusion(self, rounds: int, seed: int = 0, use_multithread: bool = False, num_threads: int = 0) -> float: ...



from pynetim.cpp.graph import IMGraphCpp
from pynetim.cpp.diffusion_model import IndependentCascadeModel, LinearThresholdModel

def test_pyi_hints():
    print("=" * 60)
    print("Testing .pyi Type Hints")
    print("=" * 60)
    
    num_nodes = 100
    num_edges = 500
    
    import random
    random.seed(42)
    
    edges = []
    weights = []
    for _ in range(num_edges):
        u = random.randint(0, num_nodes - 1)
        v = random.randint(0, num_nodes - 1)
        if u != v:
            edges.append((u, v))
            weights.append(random.random())
    
    print(f"\nCreating graph with {num_nodes} nodes and {len(edges)} edges...")
    graph = IMGraphCpp(num_nodes=num_nodes, edges=edges, weights=weights, directed=True)
    print(f"Graph: {graph}")
    
    print("\n" + "=" * 60)
    print("Testing IMGraphCpp methods with type hints")
    print("=" * 60)
    
    test_nodes = [0, 1, 2, 3, 4]
    
    print("\n1. Testing batch_out_degree...")
    out_degrees = graph.batch_out_degree(test_nodes)
    print(f"   Result: {out_degrees}")
    print(f"   Type: {type(out_degrees)}")
    
    print("\n2. Testing batch_in_degree...")
    in_degrees = graph.batch_in_degree(test_nodes)
    print(f"   Result: {in_degrees}")
    print(f"   Type: {type(in_degrees)}")
    
    print("\n3. Testing batch_degree...")
    degrees = graph.batch_degree(test_nodes)
    print(f"   Result: {degrees}")
    print(f"   Type: {type(degrees)}")
    
    print("\n4. Testing batch_out_neighbors...")
    neighbors = graph.batch_out_neighbors(test_nodes)
    print(f"   Result length: {len(neighbors)}")
    print(f"   Type: {type(neighbors)}")
    
    print("\n5. Testing batch_get_edge_weight...")
    test_edges = [(0, 1), (1, 2), (2, 3)]
    edge_weights = graph.batch_get_edge_weight(test_edges)
    print(f"   Result: {edge_weights}")
    print(f"   Type: {type(edge_weights)}")
    
    print("\n" + "=" * 60)
    print("Testing IndependentCascadeModel with type hints")
    print("=" * 60)
    
    seeds = {0, 1}
    ic_model = IndependentCascadeModel(graph, seeds)
    
    print("\n1. Testing run_monte_carlo_diffusion...")
    result = ic_model.run_monte_carlo_diffusion(rounds=100, seed=42, use_multithread=False, num_threads=0)
    print(f"   Result: {result}")
    print(f"   Type: {type(result)}")
    
    print("\n2. Testing set_seeds...")
    new_seeds = {2, 3}
    ic_model.set_seeds(new_seeds)
    print(f"   Seeds set successfully")
    
    print("\n" + "=" * 60)
    print("Testing LinearThresholdModel with type hints")
    print("=" * 60)
    
    lt_model = LinearThresholdModel(graph, seeds)
    
    print("\n1. Testing run_monte_carlo_diffusion...")
    result = lt_model.run_monte_carlo_diffusion(rounds=100, seed=42, use_multithread=False, num_threads=0)
    print(f"   Result: {result}")
    print(f"   Type: {type(result)}")
    
    print("\n2. Testing set_seeds...")
    lt_model.set_seeds(new_seeds)
    print(f"   Seeds set successfully")
    
    print("\n" + "=" * 60)
    print("All .pyi type hint tests passed!")
    print("=" * 60)

if __name__ == "__main__":
    test_pyi_hints()



# -*- coding: utf-8 -*-
__version__ = "0.4.0"
__author__ = "Zhang Kaijing"

# 1. 导入子包（让用户能访问 im.py 和 im.cpp）
from . import py
from . import cpp

__all__ = [
    'py',      # 导出子包本身
    'cpp',     # 导出子包本身
]


import heapq
import time
from collections import defaultdict

from ..graph import IMGraph
from .base_algorithm import BaseAlgorithm


class SingleDiscountAlgorithm(BaseAlgorithm):
    """
    简单度折扣启发式算法(Single Discount)用于选择影响力最大化种子节点。

    该算法通过逐步选择具有最高度数的节点作为种子，并对其邻居节点的度数进行折扣，
    以避免选择过多相互连接的节点。

    参考文献:
        - Chen, W., Wang, Y., & Yang, S. (2009). "Efficient influence maximization in social networks."
          Proceedings of the 15th ACM SIGKDD International Conference on Knowledge Discovery and Data Mining (KDD), 199-208.
          DOI: 10.1145/1557019.1557047
          URL: https://dl.acm.org/doi/10.1145/1557019.1557047
    """

    def __init__(self, graph: IMGraph, diffusion_model=None):
        """
        初始化简单度折扣算法实例。

        Args:
            graph (IMGraph): 输入图对象
            diffusion_model: 扩散模型，此算法不使用该参数
        """
        super(SingleDiscountAlgorithm, self).__init__(graph, diffusion_model)
    
    def run(self, k: int):
        """
        运行简单度折扣算法选择k个种子节点。

        Args:
            k (int): 需要选择的种子节点数量

        Returns:
            list: 选择的种子节点列表
        """
        d = dict(self.graph.out_degree())  # 节点的初始度数
        seeds = []  # 存储种子节点
        selected = set()  # 记录已选择的种子节点

        # 使用最大堆存储 (-度数, 节点)，堆中的节点按度数排序
        heap = [(-d[v], v) for v in self.graph.nodes]
        heapq.heapify(heap)

        while len(seeds) < k:
            # 选择度数最大的节点作为种子节点
            _, u = heapq.heappop(heap)

            if u in selected:
                continue
            seeds.append(u)  # 将 u 添加到种子集合中
            selected.add(u)  # 标记 u 为已选择的节点

            # 对 u 的邻居节点的度数进行折扣
            for v in self.graph.out_neighbors(u):
                if v not in selected:
                    d[v] -= 1  # 将邻居节点 v 的度数减一
                    # 重新将折扣后的节点放回堆中（更新度数）
                    heapq.heappush(heap, (-d[v], v))

        self.seeds = seeds
        return seeds


class DegreeDiscountAlgorithm(BaseAlgorithm):
    """
    度折扣启发式算法(Degree Discount)用于选择影响力最大化种子节点。

    该算法是Single Discount的改进版本，考虑了邻居节点之间的影响关系，
    使用更复杂的折扣公式来更好地评估节点的边际影响力。

    参考文献:
        - Chen, W., Wang, Y., & Yang, S. (2009). "Efficient influence maximization in social networks."
          Proceedings of the 15th ACM SIGKDD International Conference on Knowledge Discovery and Data Mining (KDD), 199-208.
          DOI: 10.1145/1557019.1557047
          URL: https://dl.acm.org/doi/10.1145/1557019.1557047
    """

    def __init__(self, graph: IMGraph, diffusion_model='IC'):
        """
        初始化度折扣算法实例。

        Args:
            graph (IMGraph): 输入图对象
            diffusion_model (str, optional): 扩散模型，默认为'IC'
        """
        super(DegreeDiscountAlgorithm, self).__init__(graph, diffusion_model)

    def run(self, k: int):
        """
        运行度折扣算法选择k个种子节点。

        Args:
            k (int): 需要选择的种子节点数量

        Returns:
            list: 选择的种子节点列表
        """
        d = dict(self.graph.out_degree())  # 节点的度
        dd = d.copy()  # 折扣度 degree discount
        t = defaultdict(int)  # t[v]: 节点v已被选为邻居种子的数量
        seeds = []  # 最终种子集合

        # 使用最大堆存储 (-折扣度, 节点)，注意 heapq 是最小堆，所以取负值
        heap = [(-dd[v], v) for v in self.graph.nodes]
        heapq.heapify(heap)
        selected = set()

        while len(seeds) < k:
            _, u = heapq.heappop(heap)
            if u in selected:
                continue
            seeds.append(u)
            selected.add(u)

            for v in self.graph.out_neighbors(u):
                if v in selected:
                    continue
                t[v] += 1
                # 获取边的权重
                weight = self.graph.edges[u, v]["weight"]
                # 更新折扣度
                dd[v] = d[v] - 2 * t[v] - (d[v] - t[v]) * t[v] * weight
                heapq.heappush(heap, (-dd[v], v))  # 更新堆中的节点

        self.seeds = seeds
        return seeds



"""
PyNetIM 测试套件

该文件包含 PyNetIM 库的完整测试用例，涵盖了：
- NetworkX 图的基本操作
- 图统计信息计算
- 图密度计算
- 图连通性分析
- 工具函数测试

运行方式:
    conda run -n pynetim python tests/test.py
"""

import sys
import networkx as nx
from pynetim.utils import graph_statistics, graph_density, connectivity_analysis, set_edge_weight, infection_threshold


def test_basic_graph_operations():
    """测试基本的图操作"""
    print("=" * 60)
    print("测试 1: 基本图操作")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.DiGraph()
    for _ in range(500):
        u = random.randint(0, 99)
        v = random.randint(0, 99)
        if u != v:
            nx_graph.add_edge(u, v)
    
    print(f"创建的图: {nx_graph.number_of_nodes()} 个节点, {nx_graph.number_of_edges()} 条边")


def test_graph_statistics():
    """测试图统计信息函数"""
    print("\n" + "=" * 60)
    print("测试 2: 图统计信息")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.DiGraph()
    for _ in range(500):
        u = random.randint(0, 99)
        v = random.randint(0, 99)
        if u != v:
            nx_graph.add_edge(u, v)
    
    print("\nNetworkX 图:")
    stats = graph_statistics(nx_graph)
    for key, value in stats.items():
        print(f"  {key}: {value}")


def test_graph_density():
    """测试图密度计算函数"""
    print("\n" + "=" * 60)
    print("测试 3: 图密度计算")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.DiGraph()
    for _ in range(500):
        u = random.randint(0, 99)
        v = random.randint(0, 99)
        if u != v:
            nx_graph.add_edge(u, v)
    
    print("\nNetworkX Graph:")
    density = graph_density(nx_graph)
    print(f"  密度: {density:.6f}")


def test_connectivity_analysis():
    """测试连通性分析函数"""
    print("\n" + "=" * 60)
    print("测试 4: 连通性分析")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.DiGraph()
    for _ in range(500):
        u = random.randint(0, 99)
        v = random.randint(0, 99)
        if u != v:
            nx_graph.add_edge(u, v)
    
    print("\nNetworkX Graph:")
    connectivity = connectivity_analysis(nx_graph)
    for key, value in connectivity.items():
        print(f"  {key}: {value}")


def test_edge_weight_setting():
    """测试边权重设置"""
    print("\n" + "=" * 60)
    print("测试 5: 边权重设置")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.DiGraph()
    for _ in range(500):
        u = random.randint(0, 99)
        v = random.randint(0, 99)
        if u != v:
            nx_graph.add_edge(u, v)
    
    print("\n测试 CONSTANT 权重:")
    set_edge_weight(nx_graph, 'CONSTANT', constant_weight=1.0)
    print(f"  边权重: {[(u, v, nx_graph.edges[(u, v)]['weight']) for u, v in list(nx_graph.edges())[:5]]}")
    
    print("\n测试 TV 权重:")
    set_edge_weight(nx_graph, 'TV')
    print(f"  边权重: {[(u, v, nx_graph.edges[(u, v)]['weight']) for u, v in list(nx_graph.edges())[:5]]}")
    
    print("\n测试 WC 权重:")
    set_edge_weight(nx_graph, 'WC')
    print(f"  边权重: {[(u, v, nx_graph.edges[(u, v)]['weight']) for u, v in list(nx_graph.edges())[:5]]}")


def test_infection_threshold():
    """测试感染阈值计算"""
    print("\n" + "=" * 60)
    print("测试 6: 感染阈值计算")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.DiGraph()
    for _ in range(500):
        u = random.randint(0, 99)
        v = random.randint(0, 99)
        if u != v:
            nx_graph.add_edge(u, v)
    
    print("\nNetworkX Graph:")
    threshold = infection_threshold(nx_graph)
    print(f"  感染阈值: {threshold:.6f}")


def run_all_tests():
    """运行所有测试"""
    print("\n" + "=" * 60)
    print("开始运行 PyNetIM 测试套件")
    print("=" * 60)
    print()
    
    try:
        test_basic_graph_operations()
        test_graph_statistics()
        test_graph_density()
        test_connectivity_analysis()
        test_edge_weight_setting()
        test_infection_threshold()
        
        print("\n" + "=" * 60)
        print("✅ 所有测试通过！")
        print("=" * 60)
        
    except Exception as e:
        print("\n" + "=" * 60)
        print(f"❌ 测试失败: {e}")
        print("=" * 60)
        sys.exit(1)


if __name__ == "__main__":
    run_all_tests()


import math
import random
import logging
from typing import List, Set, Dict, Callable, Tuple
from multiprocessing import Pool, cpu_count

from ..graph import IMGraph
from .base_algorithm import BaseAlgorithm

# 配置日志
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)


class BaseRISAlgorithm(BaseAlgorithm):
    """
    最简版反向影响采样算法(Reverse Influence Sampling, RIS)

    说明:
    - 不自动估计采样数量
    - 用户指定RR集合数量(num_rr_sets)
    - 使用堆优化贪心选择(O(R log k)而非O(R k))
    - 支持IC/LT模型

    参考文献:
        - Borgs, C., Brautbar, M., Chitnis, N., & Tardos, É. (2014). "Maximizing social influence in nearly optimal time."
          Proceedings of the 25th ACM-SIAM Symposium on Discrete Algorithms (SODA), 946-957.
          DOI: 10.1137/1.9781611973402.70
          URL: https://epubs.siam.org/doi/10.1137/1.9781611973402.70

    Attributes:
        graph (IMGraph): 输入图对象
        model (str): 扩散模型名称('IC'或'LT')
        nodes (list): 图中所有节点的列表
        rr_func (Callable): 用于生成RR集合的函数
        multi_process (bool): 是否启用多进程模式
        processes (int): 多进程模式下的进程数
        seed (int): 随机种子
    """

    def __init__(self, graph: IMGraph, diffusion_model: str = "IC", multi_process: bool = False, processes: int = None,
                 seed: int = None):
        """
        初始化基础RIS算法实例。

        Args:
            graph (IMGraph): 输入图对象
            diffusion_model (str): 扩散模型，支持'IC'或'LT'，默认为'IC'
            multi_process (bool, optional): 是否启用多进程模式，默认为False
            processes (int, optional): 多进程模式下的进程数，为None时使用CPU核心数
            seed (int, optional): 随机种子，默认为None
        """
        super(BaseRISAlgorithm, self).__init__(graph, diffusion_model)
        self.graph = graph
        self.model = diffusion_model.upper()
        self.nodes = list(graph.nodes)
        if self.model == "IC":
            self.rr_func = self._sample_rr_ic
        elif self.model == "LT":
            self.rr_func = self._sample_rr_lt
        else:
            raise ValueError("模型必须是'IC'或'LT'")

        if multi_process:
            if processes is None:
                processes = cpu_count()
            self.processes = processes
        self.multi_process = multi_process

        self.seed = seed
        random.seed(self.seed)

    def _sample_rr_ic(self, start_node: int) -> Set[int]:
        """
        IC模型：从起始节点反向BFS采样可达节点。

        Args:
            start_node (int): 起始节点

        Returns:
            Set[int]: 反向可达节点集合
        """
        active = {start_node}
        queue = [start_node]

        while queue:
            u = queue.pop(0)
            for v in self.graph.in_neighbors(u):
                if v not in active and random.random() <= self.graph.edges[v, u]['weight']:
                    active.add(v)
                    queue.append(v)
        return active

    def _sample_rr_lt(self, start_node: int) -> Set[int]:
        """
        LT模型：从起始节点反向随机走一条路径。

        Args:
            start_node (int): 起始节点

        Returns:
            Set[int]: 反向可达节点集合
        """
        active = {start_node}
        current = start_node
        nodes = list(self.graph.nodes)

        while current in nodes:
            in_neighbors = list(self.graph.in_neighbors(current))
            if not in_neighbors:
                break
            # 随机选择一个前驱（等概率）
            pred = random.choice(in_neighbors)
            if pred not in active:
                active.add(pred)
                current = pred
            else:
                break  # 遇到已激活节点，停止
        return active

    def _generate_rr_sets_single_process(self, num_rr_sets: int) -> List[Set[int]]:
        """
        在单个进程中生成RR集合。

        Args:
            num_rr_sets (int): 需要生成的RR集合数量

        Returns:
            List[Set[int]]: RR集合列表
        """
        rr_sets = []
        nodes = self.nodes
        for i in range(num_rr_sets):
            random.seed(self.seed)
            v = random.choice(nodes)
            rr_set = self.rr_func(v)
            rr_sets.append(rr_set)
        return rr_sets

    def _generate_rr_sets(self, num_rr_sets: int) -> List[Set[int]]:
        """
        生成指定数量的RR集合，支持单进程和多进程模式。

        Args:
            num_rr_sets (int): 需要生成的RR集合数量

        Returns:
            List[Set[int]]: RR集合列表
        """
        if self.multi_process:
            # 每个进程生成 num_rr_sets / processes 个RR集合
            rr_sets_per_worker = int(num_rr_sets / self.processes)
            remaining_rr_sets = num_rr_sets - rr_sets_per_worker * self.processes

            with Pool(processes=self.processes) as pool:
                # 为每个进程准备参数
                args = [(rr_sets_per_worker, ) for _ in range(self.processes)]
                # 如果有余数，分配给第一个进程处理
                if remaining_rr_sets > 0 and args:
                    args[0] = (rr_sets_per_worker + remaining_rr_sets)

                results = pool.starmap(self._generate_rr_sets_single_process, args)

            # 合并所有进程的结果
            rr_sets = []
            for result in results:
                rr_sets.extend(result)

            logger.info(f"已生成 {len(rr_sets)} 个RR集合（多进程模式）")
        else:
            # 单进程模式（原始实现）
            rr_sets = []
            for _ in range(num_rr_sets):
                v = random.choice(self.nodes)
                rr_set = self.rr_func(v)
                rr_sets.append(rr_set)
            logger.info(f"已生成 {len(rr_sets)} 个RR集合（单进程模式）")

        return rr_sets

    def _node_selection(self, rr_sets: List[Set[int]], k: int) -> Tuple[Set[int], float]:
        """
        节点选择算法，返回选出的种子集合和覆盖比例。

        Args:
            rr_sets (List[Set[int]]): RR集合列表
            k (int): 需要选择的种子节点数量

        Returns:
            Tuple[Set[int], float]: 选出的种子节点集合和覆盖比例
        """
        rr_degree = {node: 0 for node in self.graph.nodes}
        node_to_rr_idx: Dict[int, List[int]] = {}

        # 统计每个节点出现在多少个RR集合中
        for rr_idx, rr_set in enumerate(rr_sets):
            for node in rr_set:
                rr_degree[node] += 1
                node_to_rr_idx.setdefault(node, []).append(rr_idx)

        Sk = set()
        matched_count = 0

        for _ in range(k):
            # 找当前覆盖最多未覆盖RR集合的节点
            max_node = max(self.graph.nodes, key=lambda x: rr_degree[x])
            if rr_degree[max_node] == 0:
                break

            Sk.add(max_node)
            matched_count += len(node_to_rr_idx[max_node])

            # 删除已被该节点覆盖的RR集合
            for rr_idx in node_to_rr_idx[max_node]:
                for node_in_rr in rr_sets[rr_idx]:
                    rr_degree[node_in_rr] -= 1
                    node_to_rr_idx[node_in_rr].remove(rr_idx)

        coverage_ratio = matched_count / len(rr_sets) if rr_sets else 0.0
        return Sk, coverage_ratio

    def run(self, k: int, num_rr_sets: int) -> List[int]:
        """
        执行简单RIS算法。

        Args:
            k (int): 种子集合大小
            num_rr_sets (int): RR集合采样数量（越大越准）

        Returns:
            Set[int]: 选出的k个种子节点
        """
        if k <= 0 or num_rr_sets <= 0:
            return []

        logger.info(f"SimpleRIS: k={k}, num_rr_sets={num_rr_sets}, model={self.model}")

        # Step 1: 生成RR集合
        rr_sets = self._generate_rr_sets(num_rr_sets)

        # 选择节点
        seeds, _ = self._node_selection(rr_sets, k)

        return list(seeds)


class IMMAlgorithm(BaseRISAlgorithm):
    """
    基于鞅的影响力最大化算法(Influence Maximization via Martingales, IMM)

    支持IC和LT两种扩散模型

    参考文献:
        - Tang, Y., Xiao, X., & Shi, Y. (2015). "Influence maximization: Near-optimal time complexity meets practical efficiency."
          Proceedings of the 2015 ACM SIGMOD International Conference on Management of Data (SIGMOD), 75-86.
          DOI: 10.1145/2723372.2723734
          URL: https://dl.acm.org/doi/10.1145/2723372.2723734

    Attributes:
        eps (float): 近似参数ε
        l (float): 失败概率参数l
    """

    def __init__(self,
                 graph: IMGraph,
                 diffusion_model: str = "IC",
                 eps: float = 0.5,
                 l: int = 1,
                 multi_process: bool = False, processes: int = None,
                 seed: int = None):
        """
        初始化IMM算法实例。

        Args:
            graph (IMGraph): 有向带权图
            diffusion_model (str): 扩散模型，支持'IC'或'LT'，默认为'IC'
            eps (float): 近似参数ε，默认为0.5
            l (int): 失败概率参数l，默认为1
            multi_process (bool, optional): 是否启用多进程模式，默认为False
            processes (int, optional): 多进程模式下的进程数，为None时使用CPU核心数
            seed (int, optional): 随机种子，默认为None
        """
        super(IMMAlgorithm, self).__init__(graph, diffusion_model, multi_process, processes, seed)
        self.eps = eps
        self.l = l * (1 + math.log(2) / math.log(graph.number_of_nodes))

    @staticmethod
    def _log_binomial(n: int, k: int) -> float:
        """
        计算组合数的对数log(C(n,k))。

        Args:
            n (int): 总数
            k (int): 选择数

        Returns:
            float: 组合数的对数
        """
        if k == 0:
            return 0.0
        res = 0.0
        for i in range(n - k + 1, n + 1):
            res += math.log(i)
        for i in range(1, k + 1):
            res -= math.log(i)
        return res

    def _sampling(self, k: int) -> List[Set[int]]:
        """
        IMM采样主过程。

        Args:
            k (int): 需要选择的种子节点数量

        Returns:
            List[Set[int]]: 采样得到的RR集合列表
        """
        n = len(self.nodes)
        R: List[Set[int]] = []
        LB = 1.0

        eps_p = self.eps * math.sqrt(2)
        log_binom = self._log_binomial(n, k)
        lambda_p = (2 + 2 * eps_p / 3) * (
                log_binom + self.l * math.log(n) + math.log(math.log2(n))
        ) * n / (eps_p ** 2)

        # Phase 1: 二分查找下界LB
        for i in range(1, int(math.log2(n)) + 1):
            x = n / (2 ** i)
            theta_i = int(lambda_p / x)

            while len(R) < theta_i:
                v = random.choice(list(self.graph.nodes))
                R.append(self.rr_func(v))

            _, F = self._node_selection(R, k)
            if n * F >= (1 + eps_p) * x:
                LB = n * F / (1 + eps_p)
                break

        # Phase 2: 计算最终需要的采样数量theta*
        alpha = math.sqrt(self.l * math.log(n) + math.log(2))
        beta = math.sqrt((1 - 1 / math.e) * (log_binom + self.l * math.log(n) + math.log(2)))
        lambda_star = 2 * n * ((1 - 1 / math.e) * alpha + beta) ** 2 / (self.eps ** 2)
        theta_star = lambda_star / LB

        while len(R) < theta_star:
            v = random.choice(list(self.graph.nodes))
            R.append(self.rr_func(v))

        return R

    def run(self, k: int) -> List[int]:
        """
        执行IMM算法，返回大小为k的种子集合。

        Args:
            k (int): 需要选择的种子节点数量

        Returns:
            Set[int]: 选出的种子节点集合
        """
        logger.info(f"开始采样阶段...")
        R = self._sampling(k)
        logger.info(f"反向可达集数量 |R| = {len(R)}")

        logger.info(f"开始节点选择阶段...")
        seeds, _ = self._node_selection(R, k)
        # logger.info(f"选出的种子节点: {sorted(Sk)}")
        return list(seeds)



"""
PyNetIM C++ 图测试

该文件包含 PyNetIM C++ 后端的完整测试用例，涵盖了：
- C++ 图的基本操作
- 邻接查询
- 度数查询
- 边权重设置
- 扩散模型测试（IC、LT、SI、SIR）

运行方式:
    conda run -n pynetim python tests/test_cpp.py
"""

import networkx as nx
from pynetim.cpp.graph import IMGraphCpp
from pynetim.cpp.diffusion_model import (
    IndependentCascadeModel,
    LinearThresholdModel,
    SusceptibleInfectedModel,
    SusceptibleInfectedRecoveredModel
)
from pynetim.py.algorithms import DegreeDiscountAlgorithm


def test_cpp_graph_basic():
    """测试 C++ 图的基本操作"""
    print("=" * 60)
    print("测试 1: C++ 图基本操作")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.erdos_renyi_graph(100, 0.2, seed=42, directed=True)
    
    edges = list(nx_graph.edges())
    weights = [1.0 / nx_graph.in_degree(v) for u, v in nx_graph.edges()]
    
    graph = IMGraphCpp(nx_graph.number_of_nodes(), edges, weights, nx_graph.is_directed())
    print(f"C++ 图: {graph}")
    print(f"  节点数: {graph.num_nodes}")
    print(f"  边数: {graph.num_edges}")
    print(f"  是否有向: {graph.directed}")


def test_cpp_neighbor_queries():
    """测试 C++ 图的邻接查询"""
    print("\n" + "=" * 60)
    print("测试 2: C++ 图邻接查询")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.erdos_renyi_graph(100, 0.2, seed=42, directed=True)
    
    edges = list(nx_graph.edges())
    weights = [1.0 / nx_graph.in_degree(v) for u, v in nx_graph.edges()]
    
    graph = IMGraphCpp(nx_graph.number_of_nodes(), edges, weights, nx_graph.is_directed())
    
    print("\n测试节点 0-9:")
    for i in range(10):
        out_neighbors = graph.out_neighbors(i)
        in_neighbors = graph.in_neighbors(i)
        out_degree = graph.out_degree(i)
        in_degree = graph.in_degree(i)
        print(f"  节点 {i}: 出度={out_degree}, 入度={in_degree}")
        print(f"    出邻居: {[(e.to, e.weight) for e in out_neighbors]}")
        print(f"    入邻居: {list(in_neighbors)}")


def test_cpp_degree_queries():
    """测试 C++ 图的度数查询"""
    print("\n" + "=" * 60)
    print("测试 3: C++ 图度数查询")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.erdos_renyi_graph(100, 0.2, seed=42, directed=True)
    
    edges = list(nx_graph.edges())
    weights = [1.0 / nx_graph.in_degree(v) for u, v in nx_graph.edges()]
    
    graph = IMGraphCpp(nx_graph.number_of_nodes(), edges, weights, nx_graph.is_directed())
    
    print("\n批量查询度数:")
    test_nodes = list(range(0, 100, 10))
    
    out_degrees = graph.batch_out_degree(test_nodes)
    in_degrees = graph.batch_in_degree(test_nodes)
    degrees = graph.batch_degree(test_nodes)
    
    print(f"  出度: {out_degrees[:5]}")
    print(f"  入度: {in_degrees[:5]}")
    print(f"  总度: {degrees[:5]}")


def test_cpp_edge_weights():
    """测试 C++ 图的边权重设置"""
    print("\n" + "=" * 60)
    print("测试 4: C++ 图边权重设置")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.erdos_renyi_graph(100, 0.2, seed=42, directed=True)
    
    edges = list(nx_graph.edges())
    weights = [1.0 / nx_graph.in_degree(v) for u, v in nx_graph.edges()]
    
    graph = IMGraphCpp(nx_graph.number_of_nodes(), edges, weights, nx_graph.is_directed())
    
    print("\n原始边权重:")
    for i in range(min(5, len(edges))):
        u, v = edges[i]
        print(f"  ({u}, {v}): {graph.get_edge_weight(u, v):.6f}")
    
    print("\n设置 CONSTANT 权重 (1.5):")
    for u, v in edges[:10]:
        graph.update_edge_weight(u, v, 1.5)
    print(f"  设置 ({u}, {v}): 1.5")
    
    print("\n更新后的边权重:")
    for i in range(min(5, len(edges))):
        u, v = edges[i]
        print(f"  ({u}, {v}): {graph.get_edge_weight(u, v):.6f}")


def test_cpp_ic_model():
    """测试 IC 扩散模型"""
    print("\n" + "=" * 60)
    print("测试 5: IC 扩散模型")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.erdos_renyi_graph(100, 0.2, seed=42, directed=True)
    
    edges = list(nx_graph.edges())
    weights = [random.random() for _ in edges]
    
    graph = IMGraphCpp(nx_graph.number_of_nodes(), edges, weights, nx_graph.is_directed())
    
    seeds = [0, 1, 2, 3, 4]
    
    ic_model = IndependentCascadeModel(graph, seeds)
    print(f"  种子节点: {seeds}")
    
    result = ic_model.diffusion()
    print(f"  激活节点数: {len(result)}")


def test_cpp_lt_model():
    """测试 LT 扩散模型"""
    print("\n" + "=" * 60)
    print("测试 6: LT 扩散模型")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.erdos_renyi_graph(100, 0.2, seed=42, directed=True)
    
    edges = list(nx_graph.edges())
    weights = [random.random() for _ in edges]
    
    graph = IMGraphCpp(nx_graph.number_of_nodes(), edges, weights, nx_graph.is_directed())
    
    seeds = [0, 1, 2, 3, 4]
    
    lt_model = LinearThresholdModel(graph, seeds)
    print(f"  种子节点: {seeds}")
    
    result = lt_model.diffusion()
    print(f"  激活节点数: {len(result)}")


def test_cpp_si_model():
    """测试 SI 扩散模型"""
    print("\n" + "=" * 60)
    print("测试 7: SI 扩散模型")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.erdos_renyi_graph(100, 0.2, seed=42, directed=True)
    
    edges = list(nx_graph.edges())
    weights = [random.random() for _ in edges]
    
    graph = IMGraphCpp(nx_graph.number_of_nodes(), edges, weights, nx_graph.is_directed())
    
    seeds = [0]
    
    si_model = SusceptibleInfectedModel(graph, seeds, beta=0.1)
    print(f"  初始感染节点: {seeds}")
    
    result = si_model.diffusion(2)
    print(f"  感染节点数: {len(result)}")


def test_cpp_sir_model():
    """测试 SIR 扩散模型"""
    print("\n" + "=" * 60)
    print("测试 8: SIR 扩散模型")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.erdos_renyi_graph(100, 0.2, seed=42, directed=True)
    
    edges = list(nx_graph.edges())
    weights = [random.random() for _ in edges]
    
    graph = IMGraphCpp(nx_graph.number_of_nodes(), edges, weights, nx_graph.is_directed())
    
    seeds = [0]
    
    sir_model = SusceptibleInfectedRecoveredModel(graph, seeds, gamma=0.5, beta=0.1)
    print(f"  初始感染节点: {seeds}")
    
    result = sir_model.diffusion(3)
    print(f"  感染节点数: {len(result)}")


def run_all_cpp_tests():
    """运行所有 C++ 图测试"""
    print("\n" + "=" * 60)
    print("开始运行 PyNetIM C++ 图测试套件")
    print("=" * 60)
    print()
    
    try:
        test_cpp_graph_basic()
        test_cpp_neighbor_queries()
        test_cpp_degree_queries()
        test_cpp_edge_weights()
        test_cpp_ic_model()
        test_cpp_lt_model()
        test_cpp_si_model()
        test_cpp_sir_model()
        
        print("\n" + "=" * 60)
        print("✅ 所有 C++ 图测试通过！")
        print("=" * 60)
        
    except Exception as e:
        print("\n" + "=" * 60)
        print(f"❌ 测试失败: {e}")
        print("=" * 60)


if __name__ == "__main__":
    run_all_cpp_tests()


from __future__ import annotations

from ..diffusion_model import BaseDiffusionModel, IndependentCascadeModel, LinearThresholdModel
from ..graph import IMGraph


class BaseAlgorithm:
    """
    算法基类。

    该类为各种影响力最大化算法实现提供了基础框架，包含图结构和扩散模型，
    以及需要子类实现的核心方法。

    Attributes:
        graph (IMGraph): 输入图对象
        seeds (list): 种子节点集合
        diffusion_model (BaseDiffusionModel): 扩散模型实例
    """

    def __init__(self, graph: IMGraph, diffusion_model: str | BaseDiffusionModel = None):
        """
        初始化算法基类。

        Args:
            graph (IMGraph): 输入图对象
            diffusion_model (str | BaseDiffusionModel, optional): 扩散模型，支持'IC'或'LT'字符串，
                                                                 或直接传入扩散模型实例，默认为None
        """
        self.graph = graph
        self.seeds = []

        if diffusion_model is not None:
            self.diffusion_model = IndependentCascadeModel
            if isinstance(diffusion_model, str):
                if diffusion_model == 'IC':
                    self.diffusion_model = IndependentCascadeModel
                elif diffusion_model == 'LT':
                    self.diffusion_model = LinearThresholdModel
                else:
                    raise ValueError("不支持的模型：请选择 'IC' 或 'LT'")
            else:
                self.diffusion_model = diffusion_model

    def run(self, k):
        """
        执行算法的抽象方法。

        子类必须实现此方法来定义具体的算法逻辑。

        Args:
            k: 需要选择的种子节点数量

        Returns:
            算法执行结果，具体类型由子类定义

        Raises:
            NotImplementedError: 当子类未实现此方法时抛出
        """
        raise NotImplementedError



import copy
import heapq

from tqdm import tqdm

from ..graph import IMGraph
from ..diffusion_model import run_monte_carlo_diffusion, BaseDiffusionModel
from .base_algorithm import BaseAlgorithm


class GreedyAlgorithm(BaseAlgorithm):
    """
    贪婪算法 (Greedy) 选择影响力最大化种子节点

    该算法通过迭代地选择具有最大边际影响力的节点作为种子节点，
    直到选够k个种子节点。每次选择都需要计算所有候选节点的边际增益。

    Attributes:
        graph (IMGraph): 输入图对象
        diffusion_model_class (BaseDiffusionModel): 影响传播模型类
    """

    def __init__(self, graph: IMGraph, diffusion_model: BaseDiffusionModel):
        """
        初始化贪婪算法实例。

        Args:
            graph (IMGraph): 输入图对象
            diffusion_model (BaseDiffusionModel): 影响传播模型类
        """
        super().__init__(graph, diffusion_model)
        self.diffusion_model_class = diffusion_model

    def run(self, k: int, round: int, multi_process=False, processes=None, show_progress=True, seed: int = None):
        """
        运行贪婪算法选择影响力最大化种子节点。

        Args:
            k (int): 种子节点数量
            round (int): 每次计算边际增益的蒙特卡洛模拟次数
            multi_process (bool, optional): 是否启用多进程模式，默认为False
            processes (int, optional): 多进程模式下的进程数，为None时使用默认值
            show_progress (bool, optional): 是否显示进度条，默认为True
            seed (int, optional): 模拟的随机种子，默认为None

        Returns:
            list: 选择的种子节点列表
        """
        seeds = set()
        nodes = set(self.graph.nodes())
        diffusion_model = self.diffusion_model_class(self.graph, list(seeds))

        # 外层循环进度条 - 选择种子节点的进度
        outer_pbar = tqdm(range(k), desc="选择种子节点", disable=not show_progress)

        for i in outer_pbar:
            best_node = None
            best_gain = -1

            node_list = list(nodes - seeds)
            # 内层循环进度条 - 遍历候选节点的进度
            inner_pbar = tqdm(node_list, desc=f"评估候选节点({i+1}/{k})",
                             leave=False, disable=not show_progress)

            for node in inner_pbar:
                # 更新模型种子集合，重置模型
                diffusion_model.reset(list(seeds | {node}))
                avg_influence = run_monte_carlo_diffusion(diffusion_model, round, multi_process, processes, seed)
                # 更新模型种子集合，重置模型
                diffusion_model.reset(list(seeds))
                avg_without = run_monte_carlo_diffusion(diffusion_model, round, multi_process, processes, seed)
                marginal_gain = avg_influence - avg_without

                if marginal_gain > best_gain:
                    best_gain = marginal_gain
                    best_node = node

                # 在进度条上显示当前最佳边际增益
                if show_progress:
                    inner_pbar.set_postfix({'当前最佳增益': f'{best_gain:.4f}'})

            seeds.add(best_node)
            # 更新外层进度条描述信息
            if show_progress:
                outer_pbar.set_description(f"已选中节点: {best_node}")

        self.seeds = list(seeds)
        return list(seeds)


class CELFAlgorithm(BaseAlgorithm):
    """
    CELF算法选择影响力最大化种子节点

    CELF (Cost-Effective Lazy Forward) 算法是贪婪算法的优化版本，
    利用边际增益的子模特性减少计算量，通过优先队列维护节点的边际增益，
    避免重复计算已失效的边际增益。

    Attributes:
        graph (IMGraph): 输入图对象
        diffusion_model_class (BaseDiffusionModel): 影响传播模型类
    """

    def __init__(self, graph: IMGraph, diffusion_model: BaseDiffusionModel):
        """
        初始化CELF算法实例。

        Args:
            graph (IMGraph): 输入图对象
            diffusion_model (BaseDiffusionModel): 影响传播模型类
        """
        super().__init__(graph, diffusion_model)
        self.diffusion_model_class = diffusion_model

    def run(self, k: int, round: int, multi_process=False, processes=None, show_progress=True, seed: int = None):
        """
        运行CELF算法选择影响力最大化种子节点。

        Args:
            k (int): 种子节点数量
            round (int): 蒙特卡洛模拟次数
            multi_process (bool, optional): 是否启用多进程模式，默认为False
            processes (int, optional): 多进程模式下的进程数，为None时使用默认值
            show_progress (bool, optional): 是否显示进度条，默认为True
            seed (int, optional): 模拟的随机数种子，默认为None

        Returns:
            list: 选择的种子节点列表
        """
        seeds = set()
        nodes = set(self.graph.nodes())

        # Initialize marginal gains
        heap = []

        # 计算初始边际增益 - 遍历所有节点
        node_list = list(nodes)
        init_pbar = tqdm(node_list, desc="初始化边际增益", disable=not show_progress)

        diffusion_model = self.diffusion_model_class(self.graph, [])
        for node in init_pbar:
            diffusion_model.reset([node])
            mg = run_monte_carlo_diffusion(diffusion_model, round, multi_process, processes, seed)
            heap.append((-mg, node, 0))  # (负边际增益, 节点, 上次计算时种子集合大小)
        heapq.heapify(heap)

        selected = 0
        # 主循环进度条
        main_pbar = tqdm(range(k), desc="选择种子节点", disable=not show_progress)

        # 创建一个扩散模型实例用于后续计算
        diffusion_model = self.diffusion_model_class(self.graph, list(seeds))

        for _ in main_pbar:
            while True:
                neg_gain, node, last_s_size = heapq.heappop(heap)

                if last_s_size == len(seeds):
                    # 新增种子节点
                    seeds.add(node)
                    selected += 1
                    if show_progress:
                        main_pbar.set_description(f"已选中节点: {node}")
                    break
                else:
                    # 重新计算该节点边际增益
                    diffusion_model.reset(list(seeds | {node}))
                    avg_with = run_monte_carlo_diffusion(diffusion_model, round, multi_process, processes, seed)
                    diffusion_model.reset(list(seeds))
                    avg_without = run_monte_carlo_diffusion(diffusion_model, round, multi_process, processes, seed)
                    marginal_gain = avg_with - avg_without
                    heapq.heappush(heap, (-marginal_gain, node, len(seeds)))  # 修复变量名

        self.seeds = list(seeds)
        return list(seeds)



import random
from pynetim.cpp.graph import IMGraphCpp
from pynetim.cpp.diffusion_model import IndependentCascadeModel, LinearThresholdModel
from memory_monitor import MemoryMonitor, profile_memory_usage, print_memory_stats, format_memory_size

def generate_random_graph(num_nodes, num_edges, directed=True, seed=42):
    random.seed(seed)
    
    edges = []
    weights = []
    
    for _ in range(num_edges):
        u = random.randint(0, num_nodes - 1)
        v = random.randint(0, num_nodes - 1)
        if u != v:
            edges.append((u, v))
            weights.append(random.random())
    
    return IMGraphCpp(num_nodes=num_nodes, edges=edges, weights=weights, directed=directed)

def profile_graph_creation():
    print("\n" + "=" * 60)
    print("Profiling Graph Creation")
    print("=" * 60)
    
    sizes = [
        (100, 500),
        (1000, 5000),
        (5000, 25000),
        (10000, 50000),
    ]
    
    for num_nodes, num_edges in sizes:
        print(f"\nCreating graph with {num_nodes} nodes and {num_edges} edges...")
        stats = profile_memory_usage(generate_random_graph, num_nodes, num_edges)
        print_memory_stats(stats, f"Graph ({num_nodes} nodes, {num_edges} edges)")

def profile_graph_operations():
    print("\n" + "=" * 60)
    print("Profiling Graph Operations")
    print("=" * 60)
    
    graph = generate_random_graph(10000, 50000)
    print(f"Graph: {graph}")
    print(f"Baseline memory: {format_memory_size(MemoryMonitor().get_memory_usage())}")
    
    monitor = MemoryMonitor()
    
    test_nodes = list(range(0, 10000, 100))
    
    print(f"\nTesting batch_out_degree with {len(test_nodes)} nodes...")
    monitor.reset_baseline()
    degrees = graph.batch_out_degree(test_nodes)
    print(f"Memory increase: {format_memory_size(monitor.get_memory_increase())}")
    print(f"Peak memory: {format_memory_size(monitor.get_peak_memory())}")
    
    print(f"\nTesting batch_out_neighbors with {len(test_nodes)} nodes...")
    monitor.reset_baseline()
    neighbors = graph.batch_out_neighbors(test_nodes)
    print(f"Memory increase: {format_memory_size(monitor.get_memory_increase())}")
    print(f"Peak memory: {format_memory_size(monitor.get_peak_memory())}")
    
    test_edges = [(random.randint(0, 9999), random.randint(0, 9999)) for _ in range(1000)]
    
    print(f"\nTesting batch_get_edge_weight with {len(test_edges)} edges...")
    monitor.reset_baseline()
    weights = graph.batch_get_edge_weight(test_edges)
    print(f"Memory increase: {format_memory_size(monitor.get_memory_increase())}")
    print(f"Peak memory: {format_memory_size(monitor.get_peak_memory())}")

def profile_diffusion_models():
    print("\n" + "=" * 60)
    print("Profiling Diffusion Models")
    print("=" * 60)
    
    graph = generate_random_graph(10000, 50000)
    seeds = set(random.sample(range(10000), 50))
    
    print(f"\nGraph: {graph}")
    print(f"Seeds: {len(seeds)} nodes")
    print(f"Baseline memory: {format_memory_size(MemoryMonitor().get_memory_usage())}")
    
    monitor = MemoryMonitor()
    
    print(f"\nCreating IndependentCascadeModel...")
    monitor.reset_baseline()
    ic_model = IndependentCascadeModel(graph, seeds)
    print(f"Memory increase: {format_memory_size(monitor.get_memory_increase())}")
    print(f"Peak memory: {format_memory_size(monitor.get_peak_memory())}")
    
    print(f"\nRunning IC model (1000 rounds, single-thread)...")
    monitor.reset_baseline()
    result = ic_model.run_monte_carlo_diffusion(rounds=1000, use_multithread=False)
    print(f"Result: {result:.4f}")
    print(f"Memory increase: {format_memory_size(monitor.get_memory_increase())}")
    print(f"Peak memory: {format_memory_size(monitor.get_peak_memory())}")
    
    print(f"\nRunning IC model (1000 rounds, multi-thread)...")
    monitor.reset_baseline()
    result = ic_model.run_monte_carlo_diffusion(rounds=1000, use_multithread=True)
    print(f"Result: {result:.4f}")
    print(f"Memory increase: {format_memory_size(monitor.get_memory_increase())}")
    print(f"Peak memory: {format_memory_size(monitor.get_peak_memory())}")
    
    print(f"\nCreating LinearThresholdModel...")
    monitor.reset_baseline()
    lt_model = LinearThresholdModel(graph, seeds)
    print(f"Memory increase: {format_memory_size(monitor.get_memory_increase())}")
    print(f"Peak memory: {format_memory_size(monitor.get_peak_memory())}")
    
    print(f"\nRunning LT model (1000 rounds, single-thread)...")
    monitor.reset_baseline()
    result = lt_model.run_monte_carlo_diffusion(rounds=1000, use_multithread=False)
    print(f"Result: {result:.4f}")
    print(f"Memory increase: {format_memory_size(monitor.get_memory_increase())}")
    print(f"Peak memory: {format_memory_size(monitor.get_peak_memory())}")
    
    print(f"\nRunning LT model (1000 rounds, multi-thread)...")
    monitor.reset_baseline()
    result = lt_model.run_monte_carlo_diffusion(rounds=1000, use_multithread=True)
    print(f"Result: {result:.4f}")
    print(f"Memory increase: {format_memory_size(monitor.get_memory_increase())}")
    print(f"Peak memory: {format_memory_size(monitor.get_peak_memory())}")

def profile_memory_efficiency():
    print("\n" + "=" * 60)
    print("Memory Efficiency Analysis")
    print("=" * 60)
    
    sizes = [
        (1000, 5000),
        (5000, 25000),
        (10000, 50000),
    ]
    
    print("\nGraph Size vs Memory Usage:")
    print(f"{'Nodes':<10} {'Edges':<10} {'Memory (MB)':<15} {'Memory/Node (KB)':<20} {'Memory/Edge (KB)':<20}")
    print("-" * 75)
    
    for num_nodes, num_edges in sizes:
        stats = profile_memory_usage(generate_random_graph, num_nodes, num_edges)
        memory_mb = stats['memory_increase_mb']
        memory_per_node_kb = (memory_mb * 1024) / num_nodes
        memory_per_edge_kb = (memory_mb * 1024) / num_edges
        
        print(f"{num_nodes:<10} {num_edges:<10} {memory_mb:<15.2f} {memory_per_node_kb:<20.2f} {memory_per_edge_kb:<20.2f}")

def run_comprehensive_memory_profile():
    print("=" * 60)
    print("PyNetIM Memory Profiling Suite")
    print("=" * 60)
    
    profile_graph_creation()
    profile_graph_operations()
    profile_diffusion_models()
    profile_memory_efficiency()
    
    print("\n" + "=" * 60)
    print("Memory profiling completed successfully!")
    print("=" * 60)

if __name__ == "__main__":
    run_comprehensive_memory_profile()



from .base_diffusion_model import BaseDiffusionModel
from .independent_cascade_model import IndependentCascadeModel
from .linear_threshold_model import LinearThresholdModel
from .run_monte_carlo_diffusion import run_monte_carlo_diffusion
from .susceptible_infected_model import SusceptibleInfectedModel
from .susceptible_infected_recovered_model import SusceptibleInfectedRecoveredModel

__all__ = [
    'IndependentCascadeModel',
    'LinearThresholdModel',
    'run_monte_carlo_diffusion',
    'SusceptibleInfectedModel',
    'SusceptibleInfectedRecoveredModel'
]



import time
import random
import numpy as np
from pynetim.cpp.graph import IMGraphCpp
from pynetim.cpp.diffusion_model import IndependentCascadeModel, LinearThresholdModel

def generate_random_graph(num_nodes, num_edges, directed=True, seed=42):
    random.seed(seed)
    np.random.seed(seed)
    
    edges = []
    weights = []
    
    for _ in range(num_edges):
        u = random.randint(0, num_nodes - 1)
        v = random.randint(0, num_nodes - 1)
        if u != v:
            edges.append((u, v))
            weights.append(random.random())
    
    return IMGraphCpp(num_nodes=num_nodes, edges=edges, weights=weights, directed=directed)

def benchmark_graph_operations(graph, num_runs=100):
    print("\n" + "=" * 60)
    print("Graph Operations Benchmark")
    print("=" * 60)
    
    results = {}
    
    print(f"\nGraph: {graph}")
    print(f"Number of runs: {num_runs}")
    
    test_nodes = list(range(min(100, graph.num_nodes)))
    
    start_time = time.time()
    for _ in range(num_runs):
        for u in test_nodes:
            graph.out_degree(u)
    elapsed = time.time() - start_time
    results['out_degree'] = elapsed / num_runs
    print(f"out_degree (individual): {results['out_degree']*1000:.4f} ms per run")
    
    start_time = time.time()
    for _ in range(num_runs):
        graph.batch_out_degree(test_nodes)
    elapsed = time.time() - start_time
    results['batch_out_degree'] = elapsed / num_runs
    print(f"batch_out_degree: {results['batch_out_degree']*1000:.4f} ms per run")
    print(f"Speedup: {results['out_degree'] / results['batch_out_degree']:.2f}x")
    
    start_time = time.time()
    for _ in range(num_runs):
        for u in test_nodes:
            graph.out_neighbors(u)
    elapsed = time.time() - start_time
    results['out_neighbors'] = elapsed / num_runs
    print(f"\nout_neighbors (individual): {results['out_neighbors']*1000:.4f} ms per run")
    
    start_time = time.time()
    for _ in range(num_runs):
        graph.batch_out_neighbors(test_nodes)
    elapsed = time.time() - start_time
    results['batch_out_neighbors'] = elapsed / num_runs
    print(f"batch_out_neighbors: {results['batch_out_neighbors']*1000:.4f} ms per run")
    print(f"Speedup: {results['out_neighbors'] / results['batch_out_neighbors']:.2f}x")
    
    test_edges = [(random.randint(0, graph.num_nodes-1), random.randint(0, graph.num_nodes-1)) 
                 for _ in range(100)]
    
    start_time = time.time()
    for _ in range(num_runs):
        for u, v in test_edges:
            graph.get_edge_weight(u, v)
    elapsed = time.time() - start_time
    results['get_edge_weight'] = elapsed / num_runs
    print(f"\nget_edge_weight (individual): {results['get_edge_weight']*1000:.4f} ms per run")
    
    start_time = time.time()
    for _ in range(num_runs):
        graph.batch_get_edge_weight(test_edges)
    elapsed = time.time() - start_time
    results['batch_get_edge_weight'] = elapsed / num_runs
    print(f"batch_get_edge_weight: {results['batch_get_edge_weight']*1000:.4f} ms per run")
    print(f"Speedup: {results['get_edge_weight'] / results['batch_get_edge_weight']:.2f}x")
    
    return results

def benchmark_diffusion_model(model, rounds=1000, num_runs=10):
    print("\n" + "=" * 60)
    print(f"Diffusion Model Benchmark ({model.__class__.__name__})")
    print("=" * 60)
    
    results = {}
    
    print(f"\nRounds: {rounds}")
    print(f"Number of runs: {num_runs}")
    
    start_time = time.time()
    for _ in range(num_runs):
        model.run_monte_carlo_diffusion(rounds=rounds, use_multithread=False)
    elapsed = time.time() - start_time
    results['single_thread'] = elapsed / num_runs
    print(f"Single-thread: {results['single_thread']*1000:.4f} ms per run")
    
    start_time = time.time()
    for _ in range(num_runs):
        model.run_monte_carlo_diffusion(rounds=rounds, use_multithread=True)
    elapsed = time.time() - start_time
    results['multi_thread'] = elapsed / num_runs
    print(f"Multi-thread: {results['multi_thread']*1000:.4f} ms per run")
    print(f"Speedup: {results['single_thread'] / results['multi_thread']:.2f}x")
    
    return results

def run_comprehensive_benchmark():
    print("=" * 60)
    print("PyNetIM Performance Benchmark Suite")
    print("=" * 60)
    
    graph_sizes = [
        (100, 500),
        (1000, 5000),
        (5000, 25000),
    ]
    
    all_results = {}
    
    for num_nodes, num_edges in graph_sizes:
        print(f"\n\n{'='*60}")
        print(f"Testing with {num_nodes} nodes, {num_edges} edges")
        print(f"{'='*60}")
        
        graph = generate_random_graph(num_nodes, num_edges)
        graph_results = benchmark_graph_operations(graph)
        all_results[f'graph_{num_nodes}'] = graph_results
        
        seeds = set(random.sample(range(num_nodes), min(10, num_nodes)))
        
        ic_model = IndependentCascadeModel(graph, seeds)
        ic_results = benchmark_diffusion_model(ic_model, rounds=1000, num_runs=5)
        all_results[f'ic_{num_nodes}'] = ic_results
        
        lt_model = LinearThresholdModel(graph, seeds)
        lt_results = benchmark_diffusion_model(lt_model, rounds=1000, num_runs=5)
        all_results[f'lt_{num_nodes}'] = lt_results
    
    print("\n\n" + "=" * 60)
    print("Summary of Performance Improvements")
    print("=" * 60)
    
    print("\nBatch Operations Speedup:")
    for key in all_results:
        if key.startswith('graph_'):
            results = all_results[key]
            num_nodes = key.split('_')[1]
            print(f"\n{num_nodes} nodes:")
            print(f"  batch_out_degree: {results['out_degree'] / results['batch_out_degree']:.2f}x")
            print(f"  batch_out_neighbors: {results['out_neighbors'] / results['batch_out_neighbors']:.2f}x")
            print(f"  batch_get_edge_weight: {results['get_edge_weight'] / results['batch_get_edge_weight']:.2f}x")
    
    print("\n\nMulti-threading Speedup:")
    for key in all_results:
        if key.startswith('ic_'):
            results = all_results[key]
            num_nodes = key.split('_')[1]
            print(f"\nIC Model ({num_nodes} nodes): {results['single_thread'] / results['multi_thread']:.2f}x")
        
        if key.startswith('lt_'):
            results = all_results[key]
            num_nodes = key.split('_')[1]
            print(f"LT Model ({num_nodes} nodes): {results['single_thread'] / results['multi_thread']:.2f}x")
    
    print("\n" + "=" * 60)
    print("Benchmark completed successfully!")
    print("=" * 60)

if __name__ == "__main__":
    run_comprehensive_benchmark()



from .utils import (
    set_edge_weight, 
    infection_threshold,
    topk,
    truncate_padding,
    graph_statistics,
    graph_density,
    connectivity_analysis
)

__all__ = [
    'set_edge_weight', 
    'infection_threshold',
    'topk',
    'truncate_padding',
    'graph_statistics',
    'graph_density',
    'connectivity_analysis'
]



# -----------------------------
# 导入子模块对象，用户可以访问模块
# -----------------------------
from . import algorithms
from . import diffusion_model
from . import graph
from . import diffusion_model
# from . import utils

# -----------------------------
# 导入核心类到包顶层，用户可以直接访问
# -----------------------------
# from .graph import IMGraph
# from .diffusion_model import (IndependentCascadeModel, LinearThresholdModel, run_monte_carlo_diffusion,
#                               SusceptibleInfectedModel, SusceptibleInfectedRecoveredModel)

# -----------------------------
# __all__ 包含模块和顶层类
# -----------------------------
__all__ = [
    # 模块
    "algorithms",
    "diffusion_model",
    "graph",
    "diffusion_model",
    # 顶层类/函数
#     "IMGraph",
#     "IndependentCascadeModel",
#     "LinearThresholdModel",
#     "SusceptibleInfectedModel",
#     "SusceptibleInfectedRecoveredModel",
#     "run_monte_carlo_diffusion"
]



from typing import List, Tuple, Dict, Set, TYPE_CHECKING

if TYPE_CHECKING:
    pass

class Edge:
    to: int
    weight: float

class IMGraphCpp:
    num_nodes: int
    num_edges: int
    directed: bool
    edges: Dict[Tuple[int, int], float]
    
    def __init__(self, num_nodes: int, edges: List[Tuple[int, int]], weights: List[float] = ..., directed: bool = True) -> None: ...
    
    def add_edge(self, u: int, v: int, w: float = 1.0) -> None: ...
    
    def add_edges(self, edges: List[Tuple[int, int]], weights: List[float] = ...) -> None: ...
    
    def update_edge_weight(self, u: int, v: int, w: float) -> None: ...
    
    def remove_edge(self, u: int, v: int) -> None: ...
    
    def remove_edges(self, edges: List[Tuple[int, int]]) -> None: ...
    
    def out_neighbors(self, u: int) -> List[Edge]: ...
    
    def in_neighbors(self, u: int) -> List[int]: ...
    
    def out_degree(self, u: int) -> int: ...
    
    def in_degree(self, u: int) -> int: ...
    
    def degree(self, u: int) -> int: ...
    
    def get_all_degrees(self) -> List[int]: ...
    
    def get_all_in_degrees(self) -> List[int]: ...
    
    def get_all_out_degrees(self) -> List[int]: ...
    
    def batch_out_degree(self, nodes: List[int]) -> List[int]: ...
    
    def batch_in_degree(self, nodes: List[int]) -> List[int]: ...
    
    def batch_degree(self, nodes: List[int]) -> List[int]: ...
    
    def batch_out_neighbors(self, nodes: List[int]) -> List[List[Tuple[int, float]]]: ...
    
    def batch_get_edge_weight(self, edges: List[Tuple[int, int]]) -> List[float]: ...
    
    def get_adj_list(self) -> List[List[Edge]]: ...
    
    def get_adj_matrix(self) -> List[List[float]]: ...
    
    def get_adj_matrix_sparse(self) -> List[Tuple[int, int, float]]: ...
    
    def get_edge_weight(self, u: int, v: int) -> float: ...
    
    def __repr__(self) -> str: ...



import random

from .run_monte_carlo_diffusion import run_monte_carlo_diffusion
from ..graph import IMGraph
from . import BaseDiffusionModel


class LinearThresholdModel(BaseDiffusionModel):
    """
    线性阈值模型（Linear Threshold Model）。

    实现了经典的线性阈值传播模型，该模型是一种离散时间传播模型，
    每个节点都有一个随机阈值，当其已激活邻居的影响力总和超过该阈值时，
    节点被激活。每个节点对其他节点的影响力通过边权重表示。

    参考文献:
        - Granovetter, M. (1978). "Threshold models of collective behavior."
          American Journal of Sociology, 83(6), 1420-1443.
          DOI: 10.1086/226707
          URL: https://www.jstor.org/stable/2778111

    Attributes:
        activated_nodes (set): 所有已被激活的节点集合
        graph (IMGraph): 表示传播网络结构（继承自BaseDiffusionModel）
        init_seeds (list): 初始种子集（继承自BaseDiffusionModel）
        record_states (bool): 指示是否记录传播过程中的状态变化（继承自BaseDiffusionModel）
        states (list): 当record_states为True时存储传播过程的状态历史（继承自BaseDiffusionModel）
    """

    def __init__(self, graph: IMGraph, init_seeds: list, record_states: bool = False):
        """
        初始化线性阈值模型。

        Args:
            graph (IMGraph): 定义了节点和边的网络结构
            init_seeds (list): 初始激活的节点集合
            record_states (bool): 控制是否记录每一步的状态，默认为False
        """
        super(LinearThresholdModel, self).__init__(graph, init_seeds, record_states)
        self.activated_nodes = set(self.init_seeds)

    def update(self, current_activated_nodes: set):
        """
        执行一次传播更新。

        在线性阈值模型中，未激活的节点检查其已激活邻居的影响力总和，
        如果超过节点的阈值则被激活。

        Args:
            current_activated_nodes (set): 当前轮次需要尝试传播的已激活节点集合

        Returns:
            set: 本轮新激活的节点集合
        """
        # 执行一次传播过程
        new_activated_nodes = set()
        node_theta = {node: random.random() for node in range(self.graph.number_of_nodes)}
        influence_sum = {node: 0 for node in range(self.graph.number_of_nodes)}
        for node in current_activated_nodes:
            for neighbor in self.graph.neighbors(node):
                if neighbor not in self.activated_nodes:
                    influence_sum[neighbor] += self.graph.edges[node, neighbor]['weight']
                    # in_neighbors = self.graph.in_neighbors(neighbor)
                    # sum_weights = sum(self.graph.edges[in_neighbor, neighbor]['weight']
                    #                   for in_neighbor in in_neighbors
                    #                   if in_neighbor in current_activated_nodes)
                    if influence_sum[neighbor] >= node_theta[neighbor]:
                        new_activated_nodes.add(neighbor)
                        self.activated_nodes.add(neighbor)
        # 记录每轮的状态
        if self.record_states and new_activated_nodes:
            self.states.append(new_activated_nodes.copy())

        return new_activated_nodes

    def diffusion(self, update_counts: int = None):
        """
        执行完整的扩散过程。

        进行指定轮次的传播更新，直到达到最大轮次或没有新的节点被激活。

        Args:
            update_counts (int, optional): 最大更新轮次数。如果为None则持续传播直到无新节点激活

        Returns:
            set: 最终所有被激活的节点集合
        """
        if update_counts is not None and update_counts <= 0:
            raise ValueError("update_counts must be a positive integer.")
        count = 0
        current_activated_nodes = set(self.init_seeds)
        while True:
            current_activated_nodes = self.update(current_activated_nodes)
            count += 1
            if not current_activated_nodes or (update_counts and count >= update_counts):
                break
        return self.activated_nodes

    def run_monte_carlo_diffusion(self, rounds: int, update_counts: int = None, multi_process: bool = False,
                                  processes: int = None, seed: int = None):
        """
        执行蒙特卡洛模拟扩散过程。

        Args:
            rounds (int): 总模拟轮数
            update_counts (int, optional): 更新轮次数，适用于SI模型等需要限制传播轮次的模型
            multi_process (bool): 是否启用多进程模式，默认为False
            processes (int, optional): 多进程模式下的进程数，为None时使用CPU核心数
            seed (int, optional): 模拟时的随机种子

        Returns:
            float: 所有模拟轮次的平均激活节点数
        """
        return run_monte_carlo_diffusion(self, rounds, update_counts, multi_process, processes, seed)


    def reset(self, init_seeds=None):
        """
        重置模型状态。

        将模型恢复到初始状态，如果启用状态记录则清空已记录的状态历史，
        并重新设置初始种子节点状态。

        Args:
            init_seeds (list, optional): 新的初始种子节点集合，若为None则使用原有种子集
        """
        if init_seeds is None:
            init_seeds = self.init_seeds
        self.init_seeds = init_seeds
        self.activated_nodes = set(init_seeds)
        if self.record_states:
            self.states = [set(self.init_seeds)]



from __future__ import annotations

from networkx import Graph, DiGraph

from ...utils import set_edge_weight


class IMGraph:
    """
    影响最大化图类，用于封装网络图结构和边权重设置。

    该类提供了对NetworkX图对象的封装，支持设置不同类型的边权重，
    并提供了便捷的图属性访问接口。

    Attributes:
        graph (Graph | DiGraph): networkx图对象，可以是有向图或无向图
        direction (bool): 指示图是否有向
        number_of_nodes (int): 图中节点总数
        number_of_edges (int): 图中边总数
        edge_weight_type (str): 边权重类型
    """

    def __init__(self, graph: Graph | DiGraph, edge_weight_type: str, constant_weight: float = None):
        """
        初始化IMGraph实例。

        Args:
            graph (Graph | DiGraph): networkx图对象，可以是有向图或无向图
            edge_weight_type (str): 边权重类型，支持' CONSTANT '、' TV '、' WC '等模式
            constant_weight (float, optional): 当使用常量权重模式时的权重值
        """
        self.nx_graph = graph
        self.direction = graph.is_directed()
        self.number_of_nodes = graph.number_of_nodes()
        self.number_of_edges = graph.number_of_edges()
        self.edge_weight_type = edge_weight_type

        set_edge_weight(self.nx_graph, edge_weight_type, constant_weight)

    @property
    def nodes(self):
        """
        获取图中所有节点。

        Returns:
            NodeView: 图中所有节点的视图
        """
        return self.nx_graph.nodes

    @property
    def edges(self):
        """
        获取图中所有边。

        Returns:
            EdgeView: 图中所有边的视图
        """
        return self.nx_graph.edges

    def neighbors(self, node):
        """
        获取指定节点的邻居节点。

        Args:
            node: 节点标识符

        Returns:
            iterator: 指定节点的所有邻居节点迭代器
        """
        return self.nx_graph.neighbors(node)

    def in_neighbors(self, node):
        """
        获取指定节点的入邻居节点。

        对于有向图，返回前驱节点；对于无向图，返回所有邻居节点。

        Args:
            node: 节点标识符

        Returns:
            iterator: 指定节点的入邻居节点迭代器
        """
        return self.nx_graph.predecessors(node) if self.direction else self.nx_graph.neighbors(node)

    def out_neighbors(self, node):
        """
        获取指定节点的出邻居节点。（同neighbors方法）

        Args:
            node: 节点标识符

        Returns:
            iterator: 指定节点的出邻居节点迭代器
        """
        return self.nx_graph.neighbors(node)

    @property
    def in_degree(self):
        """
        获取图的入度。

        Returns:
            int: 图的入度视图
        """
        return self.nx_graph.in_degree if self.direction else self.nx_graph.degree()

    @property
    def out_degree(self):
        """
        获取图的出度。

        Returns:
            int: 图的出度视图
        """
        return self.nx_graph.out_degree if self.direction else self.nx_graph.degree()

    def degree(self):
        """
        获取图的度。

        Returns:
            int: 图的度视图
        """
        return self.nx_graph.degree()

    def batch_out_degree(self, nodes):
        """
        批量获取指定节点的出度。

        Args:
            nodes: 节点列表

        Returns:
            list: 每个节点的出度列表
        """
        if self.direction:
            return [self.nx_graph.out_degree(node) for node in nodes]
        else:
            return [self.nx_graph.degree(node) for node in nodes]

    def batch_in_degree(self, nodes):
        """
        批量获取指定节点的入度。

        Args:
            nodes: 节点列表

        Returns:
            list: 每个节点的入度列表
        """
        if self.direction:
            return [self.nx_graph.in_degree(node) for node in nodes]
        else:
            return [self.nx_graph.degree(node) for node in nodes]

    def batch_degree(self, nodes):
        """
        批量获取指定节点的度。

        Args:
            nodes: 节点列表

        Returns:
            list: 每个节点的度列表
        """
        return [self.nx_graph.degree(node) for node in nodes]

    def __str__(self):
        """
        返回图对象的字符串表示。

        Returns:
            str: 图对象的字符串描述
        """
        return f'{self.nx_graph} and edge_weight_type: {self.edge_weight_type}'

    def __repr__(self):
        """
        返回图对象的详细字符串表示。

        Returns:
            str: 图对象的详细字符串描述
        """
        return self.__str__()



import random
from typing import Union, Dict, Any, List, TYPE_CHECKING
from collections import Counter

import networkx as nx
from networkx import Graph, DiGraph

if TYPE_CHECKING:
    from pynetim.cpp.graph import IMGraphCpp

def set_edge_weight(graph: Union[Graph, DiGraph, 'IMGraphCpp'], edge_weight_type: str, constant_weight: float = None):
    """
    根据指定的权重模型为图中的边设置权重值。

    参数:
        graph (Graph|DiGraph|IMGraphCpp): 网络图对象，可以是有向图、无向图或C++图
        edge_weight_type (str): 边权重类型，支持 'CONSTANT'、'TV'、'WC' 三种模式
        constant_weight (float, optional): 当使用常量权重模式时，指定的权重值

    异常:
        ValueError: 当权重类型不支持或参数不符合要求时抛出
    """
    edge_weight_type = edge_weight_type.upper()
    
    is_cpp_graph = hasattr(graph, 'num_nodes') and hasattr(graph, 'directed')
    
    if edge_weight_type == 'CONSTANT':
        if constant_weight is None:
            raise ValueError('使用CONSTANT模型时必须提供常量权重值')
        if is_cpp_graph:
            for u, v in graph.edges.keys():
                graph.update_edge_weight(u, v, constant_weight)
        else:
            for u, v in graph.edges():
                graph.edges[(u, v)]['weight'] = constant_weight
    elif edge_weight_type == 'TV':
        weight_list = [0.001, 0.01, 0.1]
        if is_cpp_graph:
            for u, v in graph.edges.keys():
                graph.update_edge_weight(u, v, random.choice(weight_list))
        else:
            for u, v in graph.edges():
                graph.edges[(u, v)]['weight'] = random.choice(weight_list)
    elif edge_weight_type == 'WC':
        if is_cpp_graph:
            for u, v in graph.edges.keys():
                if graph.directed:
                    graph.update_edge_weight(u, v, 1.0 / graph.in_degree(v))
                else:
                    graph.update_edge_weight(u, v, 1.0 / graph.degree(v))
        else:
            for u, v in graph.edges():
                if graph.is_directed():
                    graph.edges[(u, v)]['weight'] = 1 / graph.in_degree(v)
                else:
                    graph.edges[(u, v)]['weight'] = 1 / graph.degree(v)
    else:
        raise ValueError('不支持的边权重模型')


def infection_threshold(graph: Union[Graph, DiGraph, 'IMGraphCpp']) -> float:
    """
    计算基于图graph的度分布的感染阈值。

    参数:
        graph (Graph|DiGraph|IMGraphCpp): 网络图对象，可以是有向图、无向图或C++图

    返回:
        float: 感染阈值（是1.05倍的阈值）。

    说明:
        该函数首先计算图中所有节点的度之和（k），然后计算所有节点度的平方和（k2）。
        感染阈值计算公式为 k / (k2 - k)，其中k2是度的平方和，k是度的总和。
        这个阈值在某些流行病学模型（如SIR模型）中用于预测疾病传播的临界条件。

    参考文献:
        - Pastor-Satorras, R., & Vespignani, A. (2001). "Epidemic spreading in scale-free networks."
          Physical Review Letters, 86(14), 3200-3203.
          DOI: 10.1103/PhysRevLett.86.3200
          URL: https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.86.3200

        - Pastor-Satorras, R., Castellano, C., Van Mieghem, P., & Vespignani, A. (2015). 
          "Epidemic processes in complex networks." Reviews of Modern Physics, 87(3), 925-979.
          DOI: 10.1103/RevModPhys.87.925
          URL: https://journals.aps.org/rmp/abstract/10.1103/RevModPhys.87.925
    """
    is_cpp_graph = hasattr(graph, 'num_nodes') and hasattr(graph, 'directed')
    
    if is_cpp_graph:
        degrees = graph.get_all_degrees()
    else:
        degree_dict = dict(graph.degree())
        degrees = list(degree_dict.values())
    
    k = sum(degrees)
    k2 = sum(x ** 2 for x in degrees)
    beta = k / (k2 - k)
    return beta + 0.05 * beta


def topk(res_dict: dict, k: int, largest: bool = True) -> List:
    """
    从字典中返回具有最大（或最小）k个值的键的列表。

    参数:
        res_dict (dict): 输入的字典，其值用于比较以决定键的排序。
        k (int): 要返回的键的数量。
        largest (bool): 如果为True，则返回具有最大值的k个键；如果为False，则返回具有最小值的k个键。

    返回:
        list: 一个包含k个键的列表，这些键对应于字典中最大（或最小）的值。
    """
    sorted_keys = sorted(res_dict.keys(), key=lambda x: res_dict[x], reverse=largest)
    return sorted_keys[:k]


def truncate_padding(seq, max_len, pad=0):
    if len(seq) >= max_len:
        return seq[:max_len]
    return seq + [pad] * (max_len - len(seq))


def graph_statistics(graph: Union[Graph, DiGraph, 'IMGraphCpp']) -> Dict[str, Any]:
    """
    计算并返回图的详细统计信息（支持NetworkX图和C++图）

    参数:
        graph (Graph|DiGraph|IMGraphCpp): 网络图对象，可以是有向图、无向图或C++图

    返回:
        dict: 包含以下统计信息的字典
            - num_nodes: 节点数
            - num_edges: 边数
            - avg_degree: 平均度
            - max_degree: 最大度
            - min_degree: 最小度
            - degree_distribution: 度分布
            - density: 图密度
            - is_connected: 是否连通（仅NetworkX图支持）
            - num_components: 连通分量数（仅NetworkX图支持）
    """
    is_cpp_graph = hasattr(graph, 'num_nodes') and hasattr(graph, 'directed')
    
    if is_cpp_graph:
        num_nodes = graph.num_nodes
        num_edges = graph.num_edges
        degrees = graph.get_all_degrees()
    else:
        num_nodes = graph.number_of_nodes()
        num_edges = graph.number_of_edges()
        degrees = list(dict(graph.degree()).values())
    
    if num_nodes == 0:
        return {
            'num_nodes': 0,
            'num_edges': 0,
            'avg_degree': 0.0,
            'max_degree': 0,
            'min_degree': 0,
            'degree_distribution': {},
            'density': 0.0,
            'is_connected': None,
            'num_components': None
        }
    
    avg_degree = sum(degrees) / num_nodes
    max_degree = max(degrees)
    min_degree = min(degrees)
    degree_dist = dict(Counter(degrees))
    
    if is_cpp_graph:
        is_connected = None
        num_components = None
    else:
        is_connected = nx.is_connected(graph)
        num_components = nx.number_connected_components(graph)
    
    return {
        'num_nodes': num_nodes,
        'num_edges': num_edges,
        'avg_degree': avg_degree,
        'max_degree': max_degree,
        'min_degree': min_degree,
        'degree_distribution': degree_dist,
        'density': graph_density(graph),
        'is_connected': is_connected,
        'num_components': num_components
    }


def graph_density(graph: Union[Graph, DiGraph, 'IMGraphCpp']) -> float:
    """
    计算图的密度（支持NetworkX图和C++图）

    参数:
        graph (Graph|DiGraph|IMGraphCpp): 网络图对象，可以是有向图、无向图或C++图

    返回:
        float: 图密度值，范围在[0, 1]之间

    说明:
        图密度定义为实际边数与可能的最大边数的比值
        有向图: |E| / (|V| * (|V| - 1))
        无向图: 2 * |E| / (|V| * (|V| - 1))
    """
    is_cpp_graph = hasattr(graph, 'num_nodes') and hasattr(graph, 'directed')
    
    if is_cpp_graph:
        num_nodes = graph.num_nodes
        num_edges = graph.num_edges
        is_directed = graph.directed
    else:
        num_nodes = graph.number_of_nodes()
        num_edges = graph.number_of_edges()
        is_directed = graph.is_directed()
    
    if num_nodes <= 1:
        return 0.0
    
    if is_directed:
        max_possible_edges = num_nodes * (num_nodes - 1)
    else:
        max_possible_edges = num_nodes * (num_nodes - 1) / 2.0
    
    return float(num_edges) / max_possible_edges


def connectivity_analysis(graph: Union[Graph, DiGraph, 'IMGraphCpp']) -> Dict[str, Any]:
    """
    分析图的连通性（支持NetworkX图和C++图）

    参数:
        graph (Graph|DiGraph|IMGraphCpp): 网络图对象，可以是有向图、无向图或C++图

    返回:
        dict: 包含以下连通性信息的字典
            - is_connected: 是否连通（仅NetworkX图支持）
            - num_components: 连通分量数（仅NetworkX图支持）
            - largest_component_size: 最大连通分量大小（仅NetworkX图支持）
            - component_sizes: 所有连通分量大小列表（仅NetworkX图支持）
            - weakly_connected: 是否弱连通（仅NetworkX图支持）
    """
    is_cpp_graph = hasattr(graph, 'num_nodes') and hasattr(graph, 'directed')
    
    if is_cpp_graph:
        return {
            'is_connected': None,
            'num_components': None,
            'largest_component_size': None,
            'component_sizes': None,
            'weakly_connected': None
        }
    
    num_nodes = graph.number_of_nodes()
    
    if num_nodes == 0:
        return {
            'is_connected': True,
            'num_components': 0,
            'largest_component_size': 0,
            'component_sizes': [],
            'weakly_connected': True
        }
    
    if graph.is_directed():
        components = list(nx.weakly_connected_components(graph))
        is_connected = nx.is_weakly_connected(graph)
    else:
        components = list(nx.connected_components(graph))
        is_connected = nx.is_connected(graph)
    
    component_sizes = [len(comp) for comp in components]
    largest_component_size = max(component_sizes) if component_sizes else 0
    
    return {
        'is_connected': is_connected,
        'num_components': len(components),
        'largest_component_size': largest_component_size,
        'component_sizes': component_sizes,
        'weakly_connected': is_connected
    }



from ..graph import IMGraph


class BaseDiffusionModel:
    """
    扩散模型基类。

    该类为各种扩散模型实现提供了基础框架，包含图结构、初始种子集、状态记录功能，
    以及需要子类实现的核心方法。支持在网络图上模拟信息、疾病或其他
    实体的传播过程。

    Attributes:
        graph (IMGraph): 表示传播网络结构
        init_seeds (list): 初始种子集
        record_states (bool): 指示是否记录传播过程中的状态变化
        states (list): 当record_states为True时存储传播过程的状态历史
    """

    def __init__(self, graph: IMGraph, init_seeds: list, record_states: bool = False):
        """
        初始化扩散模型基类。

        Args:
            graph (IMGraph): 定义了节点和边的网络结构
            init_seeds (list): 初始激活的节点集合
            record_states (bool): 控制是否记录每一步的状态，默认为False
        """
        self.graph = graph
        self.init_seeds = init_seeds.copy()
        self.record_states = record_states

        if self.record_states:
            self.states = [set(init_seeds)]

    def update(self):
        """
        更新模型状态的抽象方法。

        子类必须实现此方法来定义具体的更新逻辑。
        每次调用应执行一次传播步骤。

        Returns:
            本轮更新结果，具体类型由子类定义

        Raises:
            NotImplementedError: 当子类未实现此方法时抛出
        """
        raise NotImplementedError

    def diffusion(self, update_counts: int = None):
        """
        执行完整扩散过程的抽象方法。

        子类必须实现此方法来定义从开始到结束的完整传播流程。

        Args:
            update_counts (int, optional): 更新次数，控制扩散迭代的轮数

        Returns:
            扩散结果，具体类型由子类定义

        Raises:
            NotImplementedError: 当子类未实现此方法时抛出
        """
        raise NotImplementedError

    def run_monte_carlo_diffusion(self, round: int, multi_process: bool = False, processes: int = None):
        """
        执行蒙特卡洛模拟扩散过程。

        Args:
            round (int): 模拟的轮数
            multi_process (bool): 指示是否使用多进程进行模拟
            processes (int, optional): 多进程的进程数，默认为None，表示使用CPU核数

        Returns:
            模拟结果，具体类型由子类定义

        Raises:
            NotImplementedError: 当子类未实现此方法时抛出
        """
        raise NotImplementedError

    def reset(self, init_seeds=None):
        """
        重置模型状态。

        将模型恢复到初始状态，如果启用状态记录则清空已记录的状态历史，
        并重新设置初始种子节点状态。

        Args:
            init_seeds (list, optional): 新的初始种子节点集合，若为None则使用原有种子集
        """
        if init_seeds is None:
            init_seeds = self.init_seeds
        self.init_seeds = init_seeds.copy()
        if self.record_states:
            self.states = [set(self.init_seeds)]



import os
from setuptools import setup, find_packages
from pybind11.setup_helpers import Pybind11Extension, build_ext

cpp_dir = os.path.join("src", "pynetim", "cpp")

ext_modules = [
    Pybind11Extension(
        "pynetim.cpp.graph.graph",
        [os.path.join(cpp_dir, "bindings", "graph_bind.cpp")],
        include_dirs=[os.path.join(cpp_dir, "bindings"), os.path.join(cpp_dir, "include")],
        cxx_std=20,
        extra_compile_args=["-mavx2"],
    ),
    Pybind11Extension(
        "pynetim.cpp.diffusion_model.independent_cascade_model",
        [
            os.path.join(cpp_dir, "bindings", "ic_bind.cpp"),
            os.path.join(cpp_dir, "bindings", "graph_bind.cpp"),
        ],
        include_dirs=[os.path.join(cpp_dir, "bindings"), os.path.join(cpp_dir, "include")],
        cxx_std=20,
        extra_compile_args=["-mavx2"],
    ),
    Pybind11Extension(
        "pynetim.cpp.diffusion_model.linear_threshold_model",
        [
            os.path.join(cpp_dir, "bindings", "lt_bind.cpp"),
            os.path.join(cpp_dir, "bindings", "graph_bind.cpp"),
        ],
        include_dirs=[os.path.join(cpp_dir, "bindings"), os.path.join(cpp_dir, "include")],
        cxx_std=20,
        extra_compile_args=["-mavx2"],
    ),
]

setup(
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
)


import random
import networkx as nx
from pynetim.utils import graph_statistics, graph_density, connectivity_analysis

def test_graph_statistics():
    print("=" * 60)
    print("Testing graph_statistics Function")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.DiGraph()
    for _ in range(500):
        u = random.randint(0, 99)
        v = random.randint(0, 99)
        if u != v:
            nx_graph.add_edge(u, v)
    
    print("\nNetworkX Graph:")
    nx_stats = graph_statistics(nx_graph)
    for key, value in nx_stats.items():
        print(f"  {key}: {value}")
    
    print("\n" + "=" * 60)
    print("All graph_statistics tests passed!")
    print("=" * 60)

def test_graph_density():
    print("\n" + "=" * 60)
    print("Testing graph_density Function")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.DiGraph()
    for _ in range(500):
        u = random.randint(0, 99)
        v = random.randint(0, 99)
        if u != v:
            nx_graph.add_edge(u, v)
    
    print("\nNetworkX Graph:")
    nx_density = graph_density(nx_graph)
    print(f"  Density: {nx_density:.6f}")
    
    print("\n" + "=" * 60)
    print("All graph_density tests passed!")
    print("=" * 60)

def test_connectivity_analysis():
    print("\n" + "=" * 60)
    print("Testing connectivity_analysis Function")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.DiGraph()
    for _ in range(500):
        u = random.randint(0, 99)
        v = random.randint(0, 99)
        if u != v:
            nx_graph.add_edge(u, v)
    
    print("\nNetworkX Graph:")
    nx_conn = connectivity_analysis(nx_graph)
    for key, value in nx_conn.items():
        print(f"  {key}: {value}")
    
    print("\n" + "=" * 60)
    print("All connectivity_analysis tests passed!")
    print("=" * 60)

def test_disconnected_graph():
    print("\n" + "=" * 60)
    print("Testing with Disconnected Graph")
    print("=" * 60)
    
    random.seed(42)
    
    nx_graph = nx.DiGraph()
    for i in range(50):
        for j in range(i + 1, 50):
            nx_graph.add_edge(i, j)
    
    print("\nNetworkX Graph (disconnected):")
    nx_stats = graph_statistics(nx_graph)
    nx_conn = connectivity_analysis(nx_graph)
    
    print(f"  is_connected: {nx_stats['is_connected']}")
    print(f"  num_components: {nx_stats['num_components']}")
    print(f"  largest_component_size: {nx_conn['largest_component_size']}")
    print(f"  component_sizes: {nx_conn['component_sizes'][:5]}...")

def test_empty_graph():
    print("\n" + "=" * 60)
    print("Testing with Empty Graph")
    print("=" * 60)
    
    nx_graph = nx.DiGraph()
    
    print("\nEmpty NetworkX Graph:")
    nx_stats = graph_statistics(nx_graph)
    for key, value in nx_stats.items():
        print(f"  {key}: {value}")

def run_all_tests():
    print("=" * 60)
    print("Utils Module Phase 1 Tests")
    print("=" * 60)
    
    test_graph_statistics()
    test_graph_density()
    test_connectivity_analysis()
    test_disconnected_graph()
    test_empty_graph()
    
    print("\n" + "=" * 60)
    print("All Phase 1 tests passed!")
    print("=" * 60)

if __name__ == "__main__":
    run_all_tests()



import random
from multiprocessing import Pool, cpu_count
import statistics

from .base_diffusion_model import BaseDiffusionModel


def __simulate_multi_round(diffusion_model: BaseDiffusionModel, rounds: int, update_counts: int = None, seed: int = None):
    """
    在单个进程中执行多轮蒙特卡洛模拟。

    Args:
        diffusion_model (BaseDiffusionModel): 扩散模型实例
        rounds (int): 该进程需要执行的模拟轮数
        update_counts (int, optional): 更新轮次数，适用于SI/SIR等模型
        seed (int, optional): 随机种子的基础值，默认为None

    Returns:
        float: 该进程所有模拟轮次的平均激活节点数
    """
    count = 0
    for i in range(rounds):
        random.seed(seed + i if seed is not None else None)
        diffusion_model.reset()
        result = diffusion_model.diffusion(update_counts)
        count += len(result)
    return count / rounds


def run_monte_carlo_diffusion(
        diffusion_model: BaseDiffusionModel,
        rounds: int,
        update_counts: int = None,
        multi_process: bool = False,
        processes: int = None,
        seed: int = None,
):
    """
    执行蒙特卡洛模拟扩散过程，支持单进程和多进程模式。

    Args:
        diffusion_model (BaseDiffusionModel): 扩散模型实例
        rounds (int): 总模拟轮数
        update_counts (int, optional): 更新轮次数，适用于SI等模型
        multi_process (bool, optional): 是否启用多进程模式，默认为False
        processes (int, optional): 多进程模式下的进程数，为None时使用CPU核心数
        seed (int, optional): 随机种子的基础值，默认为None

    Returns:
        float: 所有模拟轮次的平均激活节点数
    """
    if rounds <= 0:
        raise ValueError("The number of rounds must be greater than 0.")

    if multi_process:
        if processes is None:
            processes = cpu_count()

        # 每个进程执行 round / processes 次模拟
        rounds_per_worker = int(rounds / processes)
        with Pool(processes=processes) as pool:
            # 每个进程需要的参数
            args = [
                (diffusion_model, rounds_per_worker, update_counts, seed + i * rounds_per_worker if seed is not None else None)
                for i in range(processes)
            ]
            results = pool.starmap(__simulate_multi_round, args)

        avg_activated = statistics.mean(results)
    else:
        # 单进程模式
        avg_activated = __simulate_multi_round(diffusion_model, rounds, update_counts, seed)

    return avg_activated



import sys
sys.path.insert(0, '/root/PyNetIM-main/src')

import time
import networkx as nx
from pynetim.cpp.graph import IMGraphCpp
from pynetim.cpp.diffusion_model import IndependentCascadeModel, LinearThresholdModel

def test_graph_basic():
    print("Testing basic Graph operations...")
    
    edges = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 0)]
    weights = [0.5, 0.3, 0.7, 0.2, 0.6]
    
    graph = IMGraphCpp(num_nodes=5, edges=edges, weights=weights, directed=True)
    
    print(f"Graph: {graph}")
    print(f"Number of nodes: {graph.num_nodes}")
    print(f"Number of edges: {graph.num_edges}")
    print(f"Directed: {graph.directed}")
    
    print("\nTesting neighbor queries...")
    for i in range(5):
        out_neighbors = graph.out_neighbors(i)
        in_neighbors = graph.in_neighbors(i)
        print(f"Node {i}: out_degree={graph.out_degree(i)}, in_degree={graph.in_degree(i)}")
        print(f"  Out neighbors: {[(e.to, e.weight) for e in out_neighbors]}")
        print(f"  In neighbors: {list(in_neighbors)}")
    
    print("\nTesting edge weight retrieval...")
    for u, v in edges:
        weight = graph.get_edge_weight(u, v)
        print(f"Edge ({u}, {v}): weight = {weight}")
    
    print("\nTesting sparse matrix...")
    sparse = graph.get_adj_matrix_sparse()
    print(f"Sparse matrix has {len(sparse)} edges")
    print(f"First 5 edges: {sparse[:5]}")
    
    print("\n✓ Basic Graph tests passed!\n")

def test_ic_model():
    print("Testing Independent Cascade Model...")
    
    edges = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 0)]
    weights = [0.5, 0.3, 0.7, 0.2, 0.6]
    
    graph = IMGraphCpp(num_nodes=5, edges=edges, weights=weights, directed=True)
    seeds = {0, 2}
    
    model = IndependentCascadeModel(graph, seeds)
    
    print(f"Seeds: {seeds}")
    
    start_time = time.time()
    result = model.run_monte_carlo_diffusion(rounds=1000, seed=42, use_multithread=False)
    single_thread_time = time.time() - start_time
    
    print(f"Single-thread result: {result:.4f} (time: {single_thread_time:.4f}s)")
    
    start_time = time.time()
    result_mt = model.run_monte_carlo_diffusion(rounds=1000, seed=42, use_multithread=True)
    multi_thread_time = time.time() - start_time
    
    print(f"Multi-thread result: {result_mt:.4f} (time: {multi_thread_time:.4f}s)")
    print(f"Speedup: {single_thread_time / multi_thread_time:.2f}x")
    
    print("\n✓ IC Model tests passed!\n")

def test_lt_model():
    print("Testing Linear Threshold Model...")
    
    edges = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 0)]
    weights = [0.5, 0.3, 0.7, 0.2, 0.6]
    
    graph = IMGraphCpp(num_nodes=5, edges=edges, weights=weights, directed=True)
    seeds = {0, 2}
    
    model = LinearThresholdModel(graph, seeds, theta_l=0.0, theta_h=1.0)
    
    print(f"Seeds: {seeds}")
    
    start_time = time.time()
    result = model.run_monte_carlo_diffusion(rounds=1000, seed=42, use_multithread=False)
    single_thread_time = time.time() - start_time
    
    print(f"Single-thread result: {result:.4f} (time: {single_thread_time:.4f}s)")
    
    start_time = time.time()
    result_mt = model.run_monte_carlo_diffusion(rounds=1000, seed=42, use_multithread=True)
    multi_thread_time = time.time() - start_time
    
    print(f"Multi-thread result: {result_mt:.4f} (time: {multi_thread_time:.4f}s)")
    print(f"Speedup: {single_thread_time / multi_thread_time:.2f}x")
    
    print("\n✓ LT Model tests passed!\n")

def test_large_graph():
    print("Testing with larger graph...")
    
    n = 1000
    p = 0.01
    
    nx_graph = nx.erdos_renyi_graph(n, p, seed=42)
    edges = list(nx_graph.edges())
    weights = [0.1] * len(edges)
    
    graph = IMGraphCpp(num_nodes=n, edges=edges, weights=weights, directed=True)
    seeds = {0, 1, 2, 3, 4}
    
    print(f"Graph: {graph}")
    
    model = IndependentCascadeModel(graph, seeds)
    
    start_time = time.time()
    result = model.run_monte_carlo_diffusion(rounds=100, seed=42, use_multithread=False)
    single_thread_time = time.time() - start_time
    
    print(f"Single-thread result: {result:.2f} (time: {single_thread_time:.4f}s)")
    
    start_time = time.time()
    result_mt = model.run_monte_carlo_diffusion(rounds=100, seed=42, use_multithread=True)
    multi_thread_time = time.time() - start_time
    
    print(f"Multi-thread result: {result_mt:.2f} (time: {multi_thread_time:.4f}s)")
    print(f"Speedup: {single_thread_time / multi_thread_time:.2f}x")
    
    print("\n✓ Large graph tests passed!\n")

if __name__ == "__main__":
    print("=" * 60)
    print("PyNetIM C++ Optimization Tests")
    print("=" * 60)
    print()
    
    test_graph_basic()
    test_ic_model()
    test_lt_model()
    test_large_graph()
    
    print("=" * 60)
    print("All tests passed successfully!")
    print("=" * 60)